# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Base classes for probability distribution objects."""

__author__ = ["fkiraly"]

__all__ = ["BaseDistribution"]

from warnings import warn

import numpy as np
import pandas as pd
from skbase.utils.dependencies import _check_estimator_deps, _check_soft_dependencies

from skpro.base import BaseObject


class BaseDistribution(BaseObject):
    """Base probability distribution."""

    # default tag values - these typically make the "safest" assumption
    _tags = {
        "object_type": "distribution",  # type of object, e.g., 'distribution'
        "python_version": None,  # PEP 440 python version specifier to limit versions
        "python_dependencies": None,  # string or str list of pkg soft dependencies
        # property tags
        # -------------
        "distr:measuretype": "mixed",  # distribution type, mixed, continuous, discrete
        "distr:paramtype": "general",
        # parameterization type - parametric, nonparametric, composite
        #
        # default parameter settings for MC estimates
        # -------------------------------------------
        # these are used in default implementations of mean, var, energy, pdfnorm, ppf
        "approx_mean_spl": 1000,  # sample size used in MC estimates of mean
        "approx_var_spl": 1000,  # sample size used in MC estimates of var
        "approx_energy_spl": 1000,  # sample size used in MC estimates of energy
        "approx_spl": 1000,  # sample size used in other MC estimates
        "bisect_iter": 1000,  # max iters for bisection method in ppf
        # which methods are approximate (not numerically exact) should be listed here
        "capabilities:approx": ["energy", "mean", "var", "pdfnorm"],
        # broadcasting and parameter settings
        # -----------------------------------
        # used to control broadcasting of parameters
        "reserved_params": ["index", "columns"],
        "broadcast_params": None,  # list of params to broadcast
        "broadcast_init": "off",  # whether to auto-broadcast params in __init__
        "broadcast_inner": "array",  # whether inner args are array or scalar-like
        # if "scalar", assumes scalar, and broadcasts in boilerplate
    }

    def __init__(self, index=None, columns=None):
        self.index = _coerce_to_pd_index_or_none(index)
        self.columns = _coerce_to_pd_index_or_none(columns)

        super().__init__()
        _check_estimator_deps(self)

        self._init_shape_bc(index=index, columns=columns)

    def _init_shape_bc(self, index=None, columns=None):
        """Initialize shape and broadcasting of distribution parameters.

        Subclasses may choose to override this, if
        default broadcasting and pre-initialization is not desired or applicable,
        e.g., distribution parameters are not array-like.

        If overridden, must set ``self._shape``: this should be an empty tuple
        if the distribution is scalar, or a pair of integers otherwise.
        """
        if self.get_tags()["broadcast_init"] == "off":
            if index is None and columns is None:
                self._shape = ()
            else:
                self._shape = (len(index), len(columns))
            return None

        # if broadcast_init os on or other, run this auto-init
        bc_params, shape, is_scalar = self._get_bc_params_dict(return_shape=True)
        self._bc_params = bc_params
        self._is_scalar = is_scalar
        self._shape = shape

        if index is None and self.ndim > 0:
            self.index = pd.RangeIndex(shape[0])

        if columns is None and self.ndim > 0:
            self.columns = pd.RangeIndex(shape[1])

    @property
    def loc(self):
        """Location indexer, for groups of indices.

        Use ``my_distribution.loc[index]`` for ``pandas``-like row/column subsetting
        of ``BaseDistribution`` descendants.

        ``index`` can be any ``pandas`` ``iloc`` compatible index subsetter.

        ``my_distribution.loc[index]``
        or ``my_distribution.loc[row_index, col_index]``
        subset ``my_distribution`` to rows selected
        by ``row_index``, cols by ``col_index``,
        to exactly the same cols/rows as ``pandas`` ``loc`` would subset
        rows in ``my_distribution.index`` and columns in ``my_distribution.columns``.
        """
        return _Indexer(ref=self, method="_loc")

    @property
    def iloc(self):
        """Integer location indexer, for groups of indices.

        Use ``my_distribution.iloc[index]`` for ``pandas``-like row/column subsetting
        of ``BaseDistribution`` descendants.

        ``index`` can be any ``pandas`` ``iloc`` compatible index subsetter.

        ``my_distribution.iloc[index]``
        or ``my_distribution.iloc[row_index, col_index]``
        subset ``my_distribution`` to rows selected
        by ``row_index``, cols by ``col_index``,
        to exactly the same cols/rows as ``pandas`` ``iloc`` would subset
        rows in ``my_distribution.index`` and columns in ``my_distribution.columns``.
        """
        return _Indexer(ref=self, method="_iloc")

    @property
    def iat(self):
        """Integer location indexer, for single index.

        Use ``my_distribution.iat[index]`` for ``pandas``-like row/column subsetting
        of ``BaseDistribution`` descendants.

        ``index`` can be any ``pandas`` ``iat`` compatible index subsetter.

        ``my_distribution.iat[index]``
        or ``my_distribution.iat[row_index, col_index]``
        subset ``my_distribution`` to the row selected
        by ``row_index``, col by ``col_index``,
        to exactly the same col/rows as ``pandas`` ``iat`` would subset
        rows in ``my_distribution.index`` and columns in ``my_distribution.columns``.
        """
        return _Indexer(ref=self, method="_iat")

    @property
    def at(self):
        """Integer location indexer, for single index.

        Use ``my_distribution.at[index]`` for ``pandas``-like row/column subsetting
        of ``BaseDistribution`` descendants.

        ``index`` can be any ``pandas`` ``at`` compatible index subsetter.

        ``my_distribution.at[index]``
        or ``my_distribution.at[row_index, col_index]``
        subset ``my_distribution`` to the row selected
        by ``row_index``, col by ``col_index``,
        to exactly the same col/rows as ``pandas`` ``at`` would subset
        rows in ``my_distribution.index`` and columns in ``my_distribution.columns``.
        """
        return _Indexer(ref=self, method="_at")

    @property
    def shape(self):
        """Shape of self, a pair (2-tuple)."""
        return self._shape

    @property
    def ndim(self):
        """Number of dimensions of self. 2 if array, 0 if scalar."""
        return len(self._shape)

    def __len__(self):
        """Length of self, number of rows."""
        shape = self._shape
        if len(shape) == 0:
            return 1
        return shape[0]

    def head(self, n=5):
        """Return the first n rows.

        If there are less than n rows in ``self``, returns clone of ``self``.

        For negative n, returns all rows except the last n.

        Parameters
        ----------
        n : int, default=5
            Number of rows to return.

        Returns
        -------
        ``self`` subset to the first n rows, i.e., ``self.iloc[0:min(n, len(self))]``
        """
        if self.ndim < 2:
            return self
        assert isinstance(n, int)
        N = len(self)
        if n < 0:
            n = N - n
        n = min(n, N)
        return self.iloc[range(n)]

    def tail(self, n=5):
        """Return the last n rows.

        If there are less than n rows in ``self``, returns clone of ``self``.

        For negative n, returns all rows except the first n.

        Parameters
        ----------
        n : int, default=5
            Number of rows to return.

        Returns
        -------
        ``self`` subset to the last n rows, i.e., ``self.iloc[max(len(self) - n, 0):]``
        """
        if self.ndim < 2:
            return self
        assert isinstance(n, int)
        N = len(self)
        if n < 0:
            start = n
        else:
            start = N - n
        start = max(0, start)
        return self.iloc[range(start, N)]

    def _loc(self, rowidx=None, colidx=None):
        if is_scalar_notnone(rowidx) and is_scalar_notnone(colidx):
            return self._at(rowidx, colidx)
        if is_scalar_notnone(rowidx):
            rowidx = pd.Index([rowidx])
        if is_scalar_notnone(colidx):
            colidx = pd.Index([colidx])

        if rowidx is not None:
            row_iloc = self.index.get_indexer_for(rowidx)
        else:
            row_iloc = None
        if colidx is not None:
            col_iloc = self.columns.get_indexer_for(colidx)
        else:
            col_iloc = None
        return self._iloc(rowidx=row_iloc, colidx=col_iloc)

    def _at(self, rowidx=None, colidx=None):
        if rowidx is not None:
            row_iloc = self.index.get_indexer_for([rowidx])[0]
        else:
            row_iloc = None
        if colidx is not None:
            col_iloc = self.columns.get_indexer_for([colidx])[0]
        else:
            col_iloc = None
        return self._iat(rowidx=row_iloc, colidx=col_iloc)

    def _subset_params(self, rowidx, colidx, coerce_scalar=False):
        """Subset distribution parameters to given rows and columns.

        Parameters
        ----------
        rowidx : None, numpy index/slice coercible, or int
            Rows to subset to. If None, no subsetting is done.
        colidx : None, numpy index/slice coercible, or int
            Columns to subset to. If None, no subsetting is done.
        coerce_scalar : bool, optional, default=False
            If True, and the subsetted parameter is a scalar, coerce it to a scalar.

        Returns
        -------
        dict
            Dictionary with subsetted distribution parameters.
            Keys are parameter names of ``self``, values are the subsetted parameters.
        """
        params = self._get_dist_params()

        subset_param_dict = {}
        for param, val in params.items():
            subset_param_dict[param] = self._subset_param(
                val=val,
                rowidx=rowidx,
                colidx=colidx,
                coerce_scalar=coerce_scalar,
            )
        return subset_param_dict

    def _subset_param(self, val, rowidx, colidx, coerce_scalar=False):
        """Subset a single distribution parameter value to given rows and columns.

        Parameters
        ----------
        val : scalar, 1D, 2D, array-like, or None
            Distribution parameter that is to be subsetted.
        rowidx : None, numpy index/slice coercible, or int
            Rows to subset to. If None, no subsetting is done.
        colidx : None, numpy index/slice coercible, or int
            Columns to subset to. If None, no subsetting is done.
        coerce_scalar : bool, optional, default=False
            If True, and the subsetted parameter is a scalar, coerce it to a scalar.

        Returns
        -------
        scalar, 1D, 2D, array-like, or None
            Subsetted distribution parameter.
        """
        if val is None:
            return None
        arr = np.array(val)
        # if len(arr.shape) == 0:
        # do nothing with arr
        if len(arr.shape) == 2 and rowidx is not None:
            arr = arr[rowidx, :]
        if len(arr.shape) == 1 and colidx is not None:
            arr = arr[colidx]
        if len(arr.shape) >= 2 and colidx is not None:
            arr = arr[:, colidx]
        if np.issubdtype(arr.dtype, np.integer):
            arr = arr.astype("float")
        if coerce_scalar:
            arr = arr[(0,) * len(arr.shape)]
        return arr

    def _iat(self, rowidx=None, colidx=None):
        if rowidx is None or colidx is None:
            raise ValueError("iat method requires both row and column index")
        subset_params = self._subset_params(
            rowidx=rowidx, colidx=colidx, coerce_scalar=True
        )
        return type(self)(**subset_params)

    def _iloc(self, rowidx=None, colidx=None):
        if is_scalar_notnone(rowidx) and is_scalar_notnone(colidx):
            return self._iat(rowidx, colidx)
        if is_scalar_notnone(rowidx):
            rowidx = pd.Index([rowidx])
        if is_scalar_notnone(colidx):
            colidx = pd.Index([colidx])

        subset_params = self._subset_params(rowidx=rowidx, colidx=colidx)

        def subset_not_none(idx, subs):
            if subs is not None:
                return idx.take(subs)
            else:
                return idx

        index_subset = subset_not_none(self.index, rowidx)
        columns_subset = subset_not_none(self.columns, colidx)

        sk_distr_type = type(self)
        return sk_distr_type(
            index=index_subset,
            columns=columns_subset,
            **subset_params,
        )

    def _get_dist_params(self):
        params = self.get_params(deep=False)
        paramnames = params.keys()
        reserved_names = ["index", "columns"]
        paramnames = [x for x in paramnames if x not in reserved_names]

        return {k: params[k] for k in paramnames}

    def get_params_df(self):
        """Return distribution parameters in a dict of DataFrame.

        Available only for simple parametric distributions,
        i.e., distributions with tag "distr:paramtype" having value "parametric".

        Returns
        -------
        dict of pd.DataFrame
            Dictionary with all distribution parameters, as ``pd.DataFrame``.
            Keys are the parameter names, values are the ``pd.DataFrame``.
            Each ``DataFrame`` has the same index as ``self`` and columns as ``self``.
            Entries are the values of the distribution parameters.
        """
        is_parametric = self.get_class_tag("distr:paramtype") == "parametric"

        if not is_parametric:
            raise RuntimeError(
                f"Error in call of {type(self).__name__}.get_params_df, "
                "DataFrame representation of parameters via get_params_df or to_df "
                "is only available for parametric distributions, i.e., "
                "distributions with tag 'distr:paramtype' being 'parametric'"
            )

        if hasattr(self, "_bc_params"):
            bc_params = self._bc_params
        else:
            bc_params = self._get_bc_params_dict()

        paramnames = list(bc_params.keys())

        def to_df(x):
            if self.ndim > 0:
                return pd.DataFrame(x, index=self.index, columns=self.columns)
            return pd.DataFrame([[x]])

        params_df = {k: to_df(bc_params[k]) for k in paramnames}
        drop_keys = ["index", "columns"]
        params_df = {k: params_df[k] for k in params_df if k not in drop_keys}
        return params_df

    def to_df(self):
        """Return distribution parameters as a single DataFrame.

        Available only for simple parametric distributions,
        i.e., distributions with tag "distr:paramtype" having value "parametric".

        Returns
        -------
        pd.DataFrame
            DataFrame with all distribution parameters.
            column is a MultiIndex (paramname, varname).
            row index is the index of the distribution.
            Entries are the values of the distribution parameters.
        """
        params_df = self.get_params_df()
        paramnames = list(params_df.keys())
        vals_df = [params_df[k] for k in paramnames]

        param_df = pd.concat(vals_df, axis=1, keys=paramnames)
        param_df.columns = param_df.columns.swaplevel()

        # sorting for consistency with columns of self
        if self.columns is not None:
            param_df = param_df.loc[:, self.columns]
        else:
            param_df = param_df.sort_index(axis=1)

        if self.ndim == 0:
            # first level is superfluous in scalar case (always 0)
            # and inconsistent with MultiIndex handling, so we remove it
            param_df = param_df.droplevel(0, axis=1)
        return param_df

    def to_str(self):
        """Return string representation of self."""
        params = self._get_dist_params()

        prt = f"{self.__class__.__name__}("
        for paramname, val in params.items():
            prt += f"{paramname}={val}, "
        prt = prt[:-2] + ")"

        return prt

    def _method_error_msg(self, method="this method", severity="warn", fill_in=None):
        msg = (
            f"{type(self)} does not have an implementation of the '{method}' method, "
            "via numerically exact implementation or fill-in approximation."
        )
        if fill_in is None:
            fill_in = "by an approximation via other methods"
        msg_approx = (
            f"{type(self)} does not have a numerically exact implementation of "
            f"the '{method}' method, it is "
            f"filled in {fill_in}."
        )
        if severity == "warn":
            return msg_approx
        else:
            return msg

    def _get_bc_params(self, *args, dtype=None, oned_as="row", return_shape=False):
        """Fully broadcast tuple of parameters given param shapes and index, columns.

        Parameters
        ----------
        args : float, int, array of floats, or array of ints (1D or 2D)
            Distribution parameters that are to be made broadcastable. If no positional
            arguments are provided, all parameters of `self` are used except for `index`
            and `columns`.
        dtype : str, optional
            broadcasted arrays are cast to all have datatype `dtype`. If None, then no
            datatype casting is done.
        oned_as : str, optional, "row" (default) or "col"
            If 'row', then 1D arrays are treated as row vectors. If 'column', then 1D
            arrays are treated as column vectors.
        return_shape : bool, optional, default=False
            If True, return shape tuple, and a boolean tuple
            indicating which parameters are scalar.

        Returns
        -------
        Tuple of float or integer arrays
            Each element of the tuple represents a different broadcastable distribution
            parameter.
        shape : Tuple, only returned if ``return_shape`` is True
            Shape of the broadcasted parameters.
            Pair of row/column if not scalar, empty tuple if scalar.
        is_scalar : Tuple of bools, only returned if ``return_is_scalar`` is True
            Each element of the tuple is True if the corresponding parameter is scalar.
        """
        number_of_params = len(args)
        if number_of_params == 0:
            # Handle case where no positional arguments are provided
            params = self._get_dist_params()
            args = tuple(params.values())
            number_of_params = len(args)

        def row_to_col(arr):
            """Convert 1D arrays to 2D col arrays, leave 2D arrays unchanged."""
            if arr.ndim == 1:
                return arr.reshape(-1, 1)
            return arr

        args_as_np = [np.array(arg) for arg in args]
        if oned_as == "col":
            args_as_np = [row_to_col(arg) for arg in args_as_np]

        if hasattr(self, "index") and self.index is not None:
            args_as_np += (self.index.to_numpy().reshape(-1, 1),)
        if hasattr(self, "columns") and self.columns is not None:
            args_as_np += (self.columns.to_numpy(),)
        bc = np.broadcast_arrays(*args_as_np)
        if dtype is not None:
            bc = [array.astype(dtype) for array in bc]
        bc = bc[:number_of_params]
        if return_shape:
            shape = bc[0].shape
            is_scalar = tuple([arr.ndim == 0 for arr in bc])
            return bc, shape, is_scalar
        return bc

    def _get_bc_params_dict(
        self, dtype=None, oned_as="row", return_shape=False, **kwargs
    ):
        """Fully broadcast dict of parameters given param shapes and index, columns.

        Parameters
        ----------
        kwargs : float, int, array of floats, or array of ints (1D or 2D)
            Distribution parameters that are to be made broadcastable. If no positional
            arguments are provided, all parameters of `self` are used except for `index`
            and `columns`.
        dtype : str, optional
            broadcasted arrays are cast to all have datatype `dtype`. If None, then no
            datatype casting is done.
        oned_as : str, optional, "row" (default) or "col"
            If 'row', then 1D arrays are treated as row vectors. If 'column', then 1D
            arrays are treated as column vectors.
        return_shape : bool, optional, default=False
            If True, return shape tuple, and a boolean tuple
            indicating which parameters are scalar.

        Returns
        -------
        dict of float or integer arrays
            Each element of the tuple represents a different broadcastable distribution
            parameter.
        shape : Tuple, only returned if ``return_shape`` is True
            Shape of the broadcasted parameters.
            Pair of row/column if not scalar, empty tuple if scalar.
        is_scalar : Tuple of bools, only returned if ``return_is_scalar`` is True
            Each element of the tuple is True if the corresponding parameter is scalar.
        """
        number_of_params = len(kwargs)
        if number_of_params == 0:
            # Handle case where no positional arguments are provided
            kwargs = self._get_dist_params()
            number_of_params = len(kwargs)

        def row_to_col(arr):
            """Convert 1D arrays to 2D col arrays, leave 2D arrays unchanged."""
            if arr.ndim == 1 and oned_as == "col":
                return arr.reshape(-1, 1)
            return arr

        kwargs_as_np = {k: row_to_col(np.array(v)) for k, v in kwargs.items()}

        if hasattr(self, "index") and self.index is not None:
            kwargs_as_np["index"] = self.index.to_numpy().reshape(-1, 1)
        if hasattr(self, "columns") and self.columns is not None:
            kwargs_as_np["columns"] = self.columns.to_numpy()

        bc_params = self.get_tags()["broadcast_params"]
        if bc_params is None:
            bc_params = kwargs_as_np.keys()
        else:
            bc_params = bc_params.copy()
            if "index" in kwargs_as_np:
                bc_params.append("index")
            if "columns" in kwargs_as_np:
                bc_params.append("columns")

        args_as_np = [kwargs_as_np[k] for k in bc_params]
        bc = np.broadcast_arrays(*args_as_np)
        if dtype is not None:
            bc = [array.astype(dtype) for array in bc]

        shape = ()
        for i, k in enumerate(bc_params):
            kwargs_as_np[k] = row_to_col(bc[i])
            if bc[i].ndim > 0:
                shape = bc[i].shape

        # special case: user provided iterables so it broadcasts to 1D
        # this is interpreted as a row vector, i.e., one multivariate distr
        if len(shape) == 1:
            shape = (1, shape[0])
            for k, v in kwargs_as_np.items():
                kwargs_as_np[k] = np.expand_dims(v, 0)

        if return_shape:
            is_scalar = tuple([arr.ndim == 0 for arr in bc])
            return kwargs_as_np, shape, is_scalar
        return kwargs_as_np

    def _boilerplate(self, method, columns=None, **kwargs):
        """Broadcasting boilerplate for distribution methods.

        Used to link public methods to private methods,
        handles coercion, broadcasting, and checks.

        Parameters
        ----------
        method : str
            Name of the method to be called, e.g., '_pdf'
        columns : None (default) or pd.Index coercible
            If not None, set return columns to this value
        kwargs : dict
            Keyword arguments to the method
            Checks and broadcasts are applied to all values in kwargs

        Examples
        --------
        >>> self._boilerplate('_pdf', x=x)  # doctest: +SKIP
        >>> # calls self._pdf(x=x_inner), broadcasting x to self's shape in x_inner
        """
        kwargs_inner = kwargs.copy()
        d = self

        for k, x in kwargs.items():
            # if x is a DataFrame, subset and reorder distribution to match it
            if isinstance(x, pd.DataFrame):
                d = self.loc[x.index, x.columns]
                x_inner = x.values
            # else, coerce to a numpy array if needed
            # then, broadcast it to the shape of self
            else:
                x_inner = self._coerce_to_self_index_np(x, flatten=False)
            kwargs_inner[k] = x_inner

        # pass the broadcasted values to the private method
        res = getattr(d, method)(**kwargs_inner)

        # if the result is not a DataFrame, coerce it to one
        # ensur the index and columns are the same as d
        if not isinstance(res, pd.DataFrame) and self.ndim > 1:
            if columns is not None:
                res_cols = pd.Index(columns)
            else:
                res_cols = d.columns
            res = pd.DataFrame(res, index=d.index, columns=res_cols)
        # if numpy scalar, convert to python scalar, e.g., float
        if isinstance(res, np.ndarray) and self.ndim == 0:
            res = res[()]
        return res

    def pdf(self, x):
        r"""Probability density function.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in ``(N, n)`` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`p_{X_{ij}}`, denote the marginal pdf of :math:`X` at the
        :math:`(i,j)`-th entry.

        The output of this method, for input ``x`` representing :math:`x`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`p_{X_{ij}}(x_{ij})`.

        If ``self`` has a mixed or discrete distribution, this returns
        the weighted continuous part of `self`'s distribution instead of the pdf,
        i.e., the marginal pdf integrate to the weight of the continuous part.

        Parameters
        ----------
        x : ``pandas.DataFrame`` or 2D ``np.ndarray``
            representing :math:`x`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`p_{X_{ij}}(x_{ij})`, as above
        """
        distr_type = self.get_tag("distr:measuretype", "mixed", raise_error=False)
        if distr_type == "discrete":
            return self._coerce_to_self_index_df(0, flatten=False)

        return self._boilerplate("_pdf", x=x)

    def _pdf(self, x):
        """Probability density function.

        Private method, to be implemented by subclasses.
        """
        self_has_logpdf = self._has_implementation_of("log_pdf")
        self_has_logpdf = self_has_logpdf or self._has_implementation_of("_log_pdf")
        if self_has_logpdf:
            approx_method = (
                "by exponentiating the output returned by the log_pdf method, "
                "this may be numerically unstable"
            )
            warn(self._method_error_msg("pdf", fill_in=approx_method))

            x = self._coerce_to_self_index_df(x, flatten=False)
            res = self.log_pdf(x=x)
            if isinstance(res, pd.DataFrame):
                res = res.values
            return np.exp(res)

        raise NotImplementedError(self._method_error_msg("pdf", "error"))

    def log_pdf(self, x):
        r"""Logarithmic probability density function.

        Numerically more stable than calling pdf and then taking logartihms.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in `(N, n)` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`p_{X_{ij}}`, denote the marginal pdf of :math:`X` at the
        :math:`(i,j)`-th entry.

        The output of this method, for input ``x`` representing :math:`x`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`\log p_{X_{ij}}(x_{ij})`.

        If ``self`` has a mixed or discrete distribution, this returns
        the weighted continuous part of `self`'s distribution instead of the pdf,
        i.e., the marginal pdf integrate to the weight of the continuous part.

        Parameters
        ----------
        x : ``pandas.DataFrame`` or 2D ``np.ndarray``
            representing :math:`x`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`\log p_{X_{ij}}(x_{ij})`, as above
        """
        distr_type = self.get_tag("distr:measuretype", "mixed", raise_error=False)
        if distr_type == "discrete":
            return self._coerce_to_self_index_df(-np.inf, flatten=False)

        return self._boilerplate("_log_pdf", x=x)

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Private method, to be implemented by subclasses.
        """
        if self._has_implementation_of("pdf") or self._has_implementation_of("_pdf"):
            approx_method = (
                "by taking the logarithm of the output returned by the pdf method, "
                "this may be numerically unstable"
            )
            warn(self._method_error_msg("log_pdf", fill_in=approx_method))

            x = self._coerce_to_self_index_df(x, flatten=False)
            res = self.pdf(x=x)
            if isinstance(res, pd.DataFrame):
                res = res.values
            return np.log(res)

        raise NotImplementedError(self._method_error_msg("log_pdf", "error"))

    def pmf(self, x):
        r"""Probability mass function.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in ``(N, n)`` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`m_{X_{ij}}`, denote the marginal mass of :math:`X` at the
        :math:`(i,j)`-th entry, i.e.,
        :math:`m_{X_{ij}}(x_{ij}) = \mathbb{P}(X_{ij} = x_{ij})`.

        The output of this method, for input ``x`` representing :math:`x`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`m_{X_{ij}}(x_{ij})`.

        If ``self`` has a mixed or discrete distribution, this returns
        the weighted continuous part of `self`'s distribution instead of the pdf,
        i.e., the marginal pdf integrate to the weight of the continuous part.

        Parameters
        ----------
        x : ``pandas.DataFrame`` or 2D ``np.ndarray``
            representing :math:`x`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`p_{X_{ij}}(x_{ij})`, as above
        """
        distr_type = self.get_tag("distr:measuretype", "mixed", raise_error=False)
        if distr_type == "continuous":
            return self._coerce_to_self_index_df(0, flatten=False)

        return self._boilerplate("_pmf", x=x)

    def _pmf(self, x):
        """Probability mass function.

        Private method, to be implemented by subclasses.
        """
        self_has_logpmf = self._has_implementation_of("log_pmf")
        self_has_logpmf = self_has_logpmf or self._has_implementation_of("_log_pmf")
        if self_has_logpmf:
            approx_method = (
                "by exponentiating the output returned by the log_pmf method, "
                "this may be numerically unstable"
            )
            warn(self._method_error_msg("pmf", fill_in=approx_method))

            x = self._coerce_to_self_index_df(x, flatten=False)
            res = self.log_pmf(x=x)
            if isinstance(res, pd.DataFrame):
                res = res.values
            return np.exp(res)

        raise NotImplementedError(self._method_error_msg("pmf", "error"))

    def log_pmf(self, x):
        r"""Logarithmic probability mass function.

        Numerically more stable than calling pmf and then taking logartihms.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in `(N, n)` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`m_{X_{ij}}`, denote the marginal pdf of :math:`X` at the
        :math:`(i,j)`-th entry, i.e.,
        :math:`m_{X_{ij}}(x_{ij}) = \mathbb{P}(X_{ij} = x_{ij})`.

        The output of this method, for input ``x`` representing :math:`x`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`\log m_{X_{ij}}(x_{ij})`.

        If ``self`` has a mixed or discrete distribution, this returns
        the weighted continuous part of `self`'s distribution instead of the pdf,
        i.e., the marginal pdf integrate to the weight of the continuous part.

        Parameters
        ----------
        x : ``pandas.DataFrame`` or 2D ``np.ndarray``
            representing :math:`x`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`\log m_{X_{ij}}(x_{ij})`, as above
        """
        distr_type = self.get_tag("distr:measuretype", "mixed", raise_error=False)
        if distr_type == "continuous":
            return self._coerce_to_self_index_df(-np.inf, flatten=False)

        return self._boilerplate("_log_pmf", x=x)

    def _log_pmf(self, x):
        """Logarithmic probability mass function.

        Private method, to be implemented by subclasses.
        """
        if self._has_implementation_of("pmf") or self._has_implementation_of("_pmf"):
            approx_method = (
                "by taking the logarithm of the output returned by the pdf method, "
                "this may be numerically unstable"
            )
            warn(self._method_error_msg("log_pmf", fill_in=approx_method))

            x = self._coerce_to_self_index_df(x, flatten=False)
            res = self.pmf(x=x)
            if isinstance(res, pd.DataFrame):
                res = res.values
            return np.log(res)

        raise NotImplementedError(self._method_error_msg("log_pmf", "error"))

    def cdf(self, x):
        r"""Cumulative distribution function.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in ``(N, n)`` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`F_{X_{ij}}`, denote the marginal cdf of :math:`X` at the
        :math:`(i,j)`-th entry,
        i.e., :math:`F_{X_{ij}}(t) = \mathbb{P}(X_{ij} \leq t)`.

        The output of this method, for input ``x`` representing :math:`x`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`F_{X_{ij}}(x_{ij})`.

        Parameters
        ----------
        x : ``pandas.DataFrame`` or 2D ``np.ndarray``
            representing :math:`x`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`F_{X_{ij}}(x_{ij})`, as above
        """
        return self._boilerplate("_cdf", x=x)

    def _cdf(self, x):
        """Cumulative distribution function.

        Private method, to be implemented by subclasses.
        """
        N = self.get_tag("approx_spl")
        approx_method = (
            "by approximating the expected value by the indicator function on "
            f"{N} samples"
        )
        warn(self._method_error_msg("mean", fill_in=approx_method))

        splx = self._sample_multiply(x, N)
        sply = self.sample(N)
        spl = splx <= sply
        return self._sample_mean(spl)

    def surv(self, x):
        r"""Survival function.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in ``(N, n)`` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`S_{X_{ij}}`, denote the marginal survival of :math:`X` at the
        :math:`(i,j)`-th entry,
        i.e., :math:`S_{X_{ij}}(t) = \mathbb{P}(X_{ij} \gneq t)`.

        The output of this method, for input ``x`` representing :math:`x`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`F_{X_{ij}}(x_{ij})`.

        Parameters
        ----------
        x : ``pandas.DataFrame`` or 2D ``np.ndarray``
            representing :math:`x`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`S_{X_{ij}}(x_{ij})`, as above
        """
        return self._boilerplate("_surv", x=x)

    def _surv(self, x):
        """Survival function.

        Private method, to be implemented by subclasses.
        """
        x = self._coerce_to_self_index_df(x, flatten=False)
        return 1 - self.cdf(x)

    def haz(self, x):
        r"""Hazard function.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in ``(N, n)`` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`h_{X_{ij}}`, denote the marginal hazard of :math:`X` at the
        :math:`(i,j)`-th entry,
        i.e., :math:`h_{X_{ij}}(t) = \frac{f_{X_{ij}}(t)}{S_{X_{ij}}(t)}`,
        where :math:`f_{X_{ij}}` is the marginal pdf, and :math:`S_{X_{ij}}`
        is the marginal survival function at the :math:`(i,j)`-th entry.

        The output of this method, for input ``x`` representing :math:`x`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`h_{X_{ij}}(x_{ij})`.

        Parameters
        ----------
        x : ``pandas.DataFrame`` or 2D ``np.ndarray``
            representing :math:`x`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`h_{X_{ij}}(x_{ij})`, as above
        """
        return self._boilerplate("_haz", x=x)

    def _haz(self, x):
        """Hazard function.

        Private method, to be implemented by subclasses.
        """
        x = self._coerce_to_self_index_df(x, flatten=False)
        return self.pdf(x) / self.surv(x)

    def ppf(self, p):
        r"""Quantile function = percent point function = inverse cdf.

        Let :math:`X` be a random variables with the distribution of ``self``,
        taking values in ``(N, n)`` ``DataFrame``-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`F_{X_{ij}}`, denote the marginal cdf of :math:`X` at the
        :math:`(i,j)`-th entry.

        The output of this method, for input ``p`` representing :math:`p`,
        is a ``DataFrame`` with same columns and indices as ``self``,
        and entries :math:`F^{-1}_{X_{ij}}(p_{ij})`.

        Parameters
        ----------
        p : ``pandas.DataFrame`` or 2D np.ndarray
            representing :math:`p`, as above

        Returns
        -------
        ``pd.DataFrame`` with same columns and index as ``self``
            containing :math:`F_{X_{ij}}(x_{ij})`, as above
        """
        return self._boilerplate("_ppf", p=p)

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        if self._has_implementation_of("cdf") or self._has_implementation_of("_cdf"):
            from scipy.optimize import bisect

            max_iter = self.get_tag("bisect_iter")
            approx_method = (
                "by using the bisection method (scipy.optimize.bisect) on "
                f"the cdf, at {max_iter} maximum iterations"
            )
            warn(self._method_error_msg("ppf", fill_in=approx_method))

            def bisect_unb(opt_fun, **kwargs):
                """Unbound version of bisect."""
                left_bd = -1e6
                right_bd = 1e6
                while opt_fun(left_bd) > 0:
                    left_bd *= 10
                while opt_fun(right_bd) < 0:
                    right_bd *= 10
                return bisect(opt_fun, left_bd, right_bd, maxiter=max_iter, **kwargs)

            shape = self.shape

            # TODO: remove duplications in the code below
            # requires cdf to accept numpy, or allow subsetting to produce scalar
            if len(shape) == 0:

                def opt_fun(x):
                    """Optimization function, to find x s.t. cdf(x) = p_ix."""
                    return self.cdf(x) - p  # noqa: B023

                result = bisect_unb(opt_fun)
                return result

            n_row, n_col = self.shape
            result = np.array([[0.0] * n_col] * n_row, dtype=float)

            for i in range(n_row):
                for j in range(n_col):
                    ix = self.index[i]
                    col = self.columns[j]
                    d_ix = self.loc[[ix], [col]]
                    p_ix = p[i, j]

                    def opt_fun(x):
                        """Optimization function, to find x s.t. cdf(x) = p_ix."""
                        x = pd.DataFrame(x, index=[ix], columns=[col])  # noqa: B023
                        return d_ix.cdf(x).values[0][0] - p_ix  # noqa: B023

                    result[i, j] = bisect_unb(opt_fun)
            return result

        raise NotImplementedError(self._method_error_msg("ppf", "error"))

    def energy(self, x=None):
        r"""Energy of self, w.r.t. self or a constant frame x.

        Let :math:`X, Y` be i.i.d. random variables with the distribution of ``self``.

        If ``x`` is ``None``, returns :math:`\mathbb{E}[|X-Y|]` (per row),
        "self-energy".
        If ``x`` is passed, returns :math:`\mathbb{E}[|X-x|]` (per row), "energy wrt x".

        The CRPS is related to energy:
        it holds that
        :math:`\mbox{CRPS}(\mbox{self}, y)` = `self.energy(y) - 0.5 * self.energy()`.

        Parameters
        ----------
        x : None or pd.DataFrame, optional, default=None
            if ``pd.DataFrame``, must have same rows and columns as ``self``

        Returns
        -------
        ``pd.DataFrame`` with same rows as ``self``, single column ``"energy"``
            each row contains one float, self-energy/energy as described above.
        """
        if x is None:
            return self._boilerplate("_energy_self", columns=["energy"])
        return self._boilerplate("_energy_x", x=x, columns=["energy"])

    def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        return self._energy_default()

    def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        return self._energy_default(x)

    def _energy_default(self, x=None):
        """Energy of self, w.r.t. self or a constant frame x.

        Default implementation, using Monte Carlo estimates.

        Parameters
        ----------
        x : None or 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        approx_spl_size = self.get_tag("approx_energy_spl")
        if x is not None and self._has_implementation_of("_ppf"):
            approx_method = (
                "by approximating the energy expectation by the integral "
                "of the absolute difference of x to the ppf,"
                f"with {approx_spl_size} equidistant nodes"
            )
            warn(self._method_error_msg("energy", fill_in=approx_method))

            ps = np.linspace(0, 1, approx_spl_size + 2)[1:-1]
            qs = [np.abs(self.ppf(p) - x) for p in ps]
            en3D = np.array(qs)
            energy = np.mean(en3D, axis=0)
            if self.ndim > 0:
                energy = np.sum(energy, axis=1)
            return energy

        # we want to approximate E[abs(X-Y)]
        # if x = None, X,Y are i.i.d. copies of self
        # if x is not None, X=x (constant), Y=self
        approx_spl_size = self.get_tag("approx_energy_spl")
        approx_method = (
            "by approximating the energy expectation by the arithmetic mean of "
            f"{approx_spl_size} samples"
        )
        warn(self._method_error_msg("energy", fill_in=approx_method))

        # splx, sply = i.i.d. samples of X - Y of size N = approx_spl_size
        N = approx_spl_size
        if x is None:
            splx = self.sample(N)
            sply = self.sample(N)
        else:  # if x is not None
            splx = self._sample_multiply(x, N)
            sply = self.sample(N)

        # approx E[abs(X-Y)] via mean of samples of abs(X-Y) obtained from splx, sply
        spl = splx - sply
        energy = spl.apply(np.linalg.norm, axis=1, ord=1)

        # todo: check if can use self._sample_mean
        if self.ndim > 0:
            energy = energy.groupby(level=1, sort=False)
        energy = energy.mean()
        if self.ndim > 0:
            energy = pd.DataFrame(energy, index=self.index, columns=["energy"])
        return energy

    def _sample_multiply(self, x, N):
        """Generate N copies of x, in a format as returned by self.sample.

        Auxiliary function used in defaults for private methods.

        Parameters
        ----------
        x : same format as output of sample(), without N,
            or np.ndarray of same shape (2D or 0D)
        N :int

        Returns
        -------
        same format as output of sample(N), containing N copies of x
        """
        if self.ndim > 0:  # and x is not None
            if not isinstance(x, pd.DataFrame):
                x = pd.DataFrame(x, index=self.index, columns=self.columns)
            spl = pd.concat([x] * N, keys=range(N))
        else:  # if self.ndim == 0 and x is not None
            spl = pd.DataFrame([x] * N)
        return spl

    def _sample_mean(self, spl):
        """Take mean of sample as returned by self.sample, respecting shape.

        Auxiliary function used in defaults for private methods.

        Parameters
        ----------
        x : same format as output of sample(N), with N:int,

        Returns
        -------
        mean of sample:
        if ``self`` is array: ``pd.DataFrame`` with same ``index`` and ``columns``
        as ``self;
        if ``self`` is scalar: scalar
        """
        if self.ndim > 0:
            levels = list(range(1, spl.index.nlevels))
            return spl.groupby(level=levels, sort=False).mean()
        else:
            return spl.mean().iloc[0]

    def mean(self):
        r"""Return expected value of the distribution.

        Let :math:`X` be a random variable with the distribution of ``self``.
        Returns the expectation :math:`\mathbb{E}[X]`

        Returns
        -------
        ``pd.DataFrame`` with same rows, columns as ``self``
            expected value of distribution (entry-wise)
        """
        return self._boilerplate("_mean")

    def _mean(self):
        """Return expected value of the distribution.

        Private method, to be implemented by subclasses.
        """
        approx_spl_size = self.get_tag("approx_mean_spl")
        if self._has_implementation_of("_ppf"):
            approx_method = (
                "by approximating the expected value by the integral of the ppf, "
                f"with {approx_spl_size} equidistant nodes"
            )
            warn(self._method_error_msg("mean", fill_in=approx_method))

            ps = np.linspace(0, 1, approx_spl_size + 2)[1:-1]
            qs = [self.ppf(p) for p in ps]
            np3D = np.array(qs)
            means = np.mean(np3D, axis=0)
            return means

        # else we have to rely on samples
        approx_method = (
            "by approximating the expected value by the arithmetic mean of "
            f"{approx_spl_size} samples"
        )
        warn(self._method_error_msg("mean", fill_in=approx_method))

        spl = self.sample(approx_spl_size)
        return self._sample_mean(spl)

    def var(self):
        r"""Return element/entry-wise variance of the distribution.

        Let :math:`X` be a random variable with the distribution of ``self``.
        Returns :math:`\mathbb{V}[X] = \mathbb{E}\left(X - \mathbb{E}[X]\right)^2`,
        where the square is element-wise.

        Returns
        -------
        ``pd.DataFrame`` with same rows, columns as ``self``
            variance of distribution (entry-wise)
        """
        return self._boilerplate("_var")

    def _var(self):
        """Return element/entry-wise variance of the distribution.

        Private method, to be implemented by subclasses.
        """
        approx_spl_size = self.get_tag("approx_var_spl")
        if self._has_implementation_of("_ppf"):
            approx_method = (
                "by approximating the variancee integrals of the ppf, "
                "integral of ppf-squared minus square of integral of ppf, "
                f"each with {approx_spl_size} equidistant nodes"
            )
            warn(self._method_error_msg("var", fill_in=approx_method))

            ps = np.linspace(0, 1, approx_spl_size + 2)[1:-1]
            qs = [self.ppf(p) for p in ps]
            qsq = [q**2 for q in qs]

            mean3D = np.array(qs)
            means = np.mean(mean3D, axis=0)

            mom2s3D = np.array(qsq)
            mom2s = np.mean(mom2s3D, axis=0)

            return mom2s - means**2

        approx_method = (
            "by approximating the variance by the arithmetic mean of "
            f"{approx_spl_size} samples of squared differences"
        )
        warn(self._method_error_msg("var", fill_in=approx_method))

        spl1 = self.sample(approx_spl_size)
        spl2 = self.sample(approx_spl_size)
        spl = (spl1 - spl2) ** 2
        return self._sample_mean(spl) / 2

    def pdfnorm(self, a=2):
        r"""a-norm of pdf, defaults to 2-norm.

        computes a-norm of the entry marginal pdf, i.e.,
        :math:`\mathbb{E}[p_X(X)^{a-1}] = \int p(x)^a dx`,
        where :math:`X` is a random variable distributed according to the entry marginal
        of `self`, and :math:`p_X` is its pdf

        Parameters
        ----------
        a: int or float, optional, default=2

        Returns
        -------
        pd.DataFrame with same rows and columns as `self`
        each entry is :math:`\mathbb{E}[p_X(X)^{a-1}] = \int p(x)^a dx`, see above
        """
        # special case: if a == 1, this is just the integral of the pdf, which is 1
        if a == 1:
            return pd.DataFrame(1.0, index=self.index, columns=self.columns)

        approx_spl_size = self.get_tag("approx_spl")
        approx_method = (
            f"by approximating the {a}-norm of the pdf by the arithmetic mean of "
            f"{approx_spl_size} samples"
        )
        warn(self._method_error_msg("pdfnorm", fill_in=approx_method))

        # uses formula int p(x)^a dx = E[p(X)^{a-1}], and MC approximates the RHS
        spl = [self.pdf(self.sample()) ** (a - 1) for _ in range(approx_spl_size)]
        spl_df = pd.concat(spl, keys=range(approx_spl_size))
        return spl_df.groupby(level=1, sort=False).mean()

    def _coerce_to_self_index_df(self, x, flatten=True):
        """Coerce input to type similar to self.

        If self is not scalar with index and columns,
        coerces x to a pd.DataFrame with index and columns as self.

        If self is scalar, coerces x to a scalar (0D) np.ndarray.
        """
        x = np.array(x)
        if flatten:
            x = x.reshape(1, -1)
        df_shape = self.shape
        x = np.broadcast_to(x, df_shape)
        if self.ndim != 0:
            df = pd.DataFrame(x, index=self.index, columns=self.columns)
            return df
        return x

    def _coerce_to_self_index_np(self, x, flatten=False):
        """Coerce input to type similar to self.

        Coerces x to a np.ndarray with same shape as self.
        Broadcasts x to self.shape, if necessary, via np.broadcast_to.

        Parameters
        ----------
        x : array-like, np.ndarray coercible
            input to be coerced to self
        flatten : bool, optional, default=True
            if True, flattens x before broadcasting
            if False, broadcasts x as is
        """
        x = np.array(x)
        if flatten:
            x = x.reshape(1, -1)
        df_shape = self.shape
        x = np.broadcast_to(x, df_shape)
        return x

    def quantile(self, alpha):
        """Return entry-wise quantiles, in Proba/pred_quantiles mtype format.

        This method broadcasts as follows:
        for a scalar `alpha`, computes the `alpha`-quantile entry-wise,
        and returns as a `pd.DataFrame` with same index, and columns as in return.
        If `alpha` is iterable, multiple quantiles will be calculated,
        and the result will be concatenated column-wise (axis=1).

        The `ppf` method also computes quantiles, but broadcasts differently, in
        `numpy` style closer to `tensorflow`.
        In contrast, this `quantile` method broadcasts
        as ``sktime`` forecaster `predict_quantiles`, i.e., columns first.

        Parameters
        ----------
        alpha : float or list of float of unique values
            A probability or list of, at which quantiles are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from `self.columns`,
            second level being the values of `alpha` passed to the function.
            Row index is `self.index`.
            Entries in the i-th row, (j, p)-the column is
            the p-th quantile of the marginal of `self` at index (i, j).
        """
        if not isinstance(alpha, list):
            alpha = [alpha]

        qdfs = []
        for p in alpha:
            p = self._coerce_to_self_index_df(p, flatten=self.ndim > 0)
            qdf = self.ppf(p)
            qdfs += [qdf]

        if self.ndim > 0:
            qres = pd.concat(qdfs, axis=1, keys=alpha)
            qres = qres.reorder_levels([1, 0], axis=1)
        else:
            qdfs = np.expand_dims(np.array(qdfs), 0)
            qres = pd.DataFrame(qdfs, columns=alpha)

        if self.ndim > 0:
            cols = pd.MultiIndex.from_product([self.columns, alpha])
        else:
            clsname = self.__class__.__name__
            cols = pd.MultiIndex.from_product([[clsname], alpha])
            qres.columns = cols

        quantiles = qres.loc[:, cols]
        return quantiles

    def sample(self, n_samples=None):
        """Sample from the distribution.

        Parameters
        ----------
        n_samples : int, optional, default = None
            number of samples to draw from the distribution

        Returns
        -------
        pd.DataFrame
            samples from the distribution

            * if ``n_samples`` is ``None``:
            returns a sample that contains a single sample from ``self``,
            in ``pd.DataFrame`` mtype format convention, with ``index`` and ``columns``
            as ``self``
            * if n_samples is ``int``:
            returns a ``pd.DataFrame`` that contains ``n_samples`` i.i.d.
            samples from ``self``, in ``pd-multiindex`` mtype format convention,
            with same ``columns`` as ``self``, and row ``MultiIndex`` that is product
            of ``RangeIndex(n_samples)`` and ``self.index``
        """

        def gen_unif():
            np_unif = np.random.uniform(size=self.shape)
            if self.ndim > 0:
                return pd.DataFrame(np_unif, index=self.index, columns=self.columns)
            return np_unif

        # if ppf is implemented, we use inverse transform sampling
        if self._has_implementation_of("_ppf") or self._has_implementation_of("ppf"):
            if n_samples is None:
                return self.ppf(gen_unif())
            # else, we generate n_samples i.i.d. samples
            pd_smpl = [self.ppf(gen_unif()) for _ in range(n_samples)]
            if self.ndim > 0:
                df_spl = pd.concat(pd_smpl, keys=range(n_samples))
            else:
                df_spl = pd.DataFrame(pd_smpl)
            return df_spl

        raise NotImplementedError(self._method_error_msg("sample", "error"))

    def plot(self, fun=None, ax=None, **kwargs):
        """Plot the distribution.

        Different distribution defining functions can be selected for plotting
        via the ``fun`` parameter.
        The functions available are the same as the methods of the distribution class,
        e.g., ``"pdf"``, ``"cdf"``, ``"ppf"``.

        For array distribution, the marginal distribution at each entry is plotted,
        as a separate subplot.

        Parameters
        ----------
        fun : str, optional, default="pdf" for continuous distributions, otherwise "cdf"
            the function to plot, one of "pdf", "cdf", "ppf"
        ax : matplotlib Axes object, optional
            matplotlib Axes to plot in
            if not provided, defaults to current axes (``plot.gca``)
        kwargs : keyword arguments
            passed to the plotting function

        Returns
        -------
        fig : matplotlib.Figure, only returned if self is array distribution
            matplotlig Figure object for subplots
        ax : matplotlib.Axes
            the axis or axes on which the plot is drawn
        """
        _check_soft_dependencies("matplotlib", obj="distribution plot")

        from matplotlib.pyplot import subplots

        if fun is None:
            if self.get_tag("distr:measuretype", "mixed") == "continuous":
                fun = "pdf"
            else:
                fun = "cdf"

        if self.ndim > 0:
            if "x_bounds" not in kwargs:
                upper = self.ppf(0.999).values.flatten().max()
                lower = self.ppf(0.001).values.flatten().min()
                x_bounds = (lower, upper)
            else:
                x_bounds = kwargs.pop("x_bounds")
            if "sharex" not in kwargs:
                sharex = True
            else:
                sharex = kwargs.pop("sharex")
            if "sharey" not in kwargs:
                sharey = True
            else:
                sharey = kwargs.pop("sharey")

            x_argname = _get_first_argname(getattr(self, fun))

            def get_ax(ax, i, j, shape):
                """Get axes at iloc i, j - API unifier for 2D and 1D subplot figures.

                Covers inconsistency in matplotlib where creation of (m, 1) matrix
                of subplots creates a 1D object and not a 2D object.
                """
                if shape[1] == 1:
                    return ax[i]
                else:
                    return ax[i, j]

            shape = self.shape
            fig, ax = subplots(shape[0], shape[1], sharex=sharex, sharey=sharey)
            for i, j in np.ndindex(shape):
                d_ij = self.iloc[i, j]
                ax_ij = get_ax(ax, i, j, shape)
                d_ij.plot(
                    fun=fun,
                    ax=ax_ij,
                    x_bounds=x_bounds,
                    print_labels="off",
                    x_argname=x_argname,
                    **kwargs,
                )
            for i in range(shape[0]):
                ax_i0 = get_ax(ax, i, 0, shape)
                ax_i0.set_ylabel(f"{self.index[i]}")
            for j in range(shape[1]):
                ax_0j = get_ax(ax, 0, j, shape)
                ax_0j.set_title(f"{self.columns[j]}")
            fig.supylabel(f"{fun}({x_argname})")
            fig.supxlabel(f"{x_argname}")
            return fig, ax

        # for now, all plots default to this function
        # but this could be changed to a dispatch mechanism
        # e.g., using this line instead
        # plot_fun_name = f"_plot_{fun}"
        plot_fun_name = "_plot_single"

        ax = getattr(self, plot_fun_name)(ax=ax, fun=fun, **kwargs)
        return ax

    def _plot_single(self, ax=None, **kwargs):
        """Plot the pdf of the distribution."""
        import matplotlib.pyplot as plt

        fun = kwargs.pop("fun")
        print_labels = kwargs.pop("print_labels", "on")
        x_argname = kwargs.pop("x_argname", "x")

        # obtain x axis bounds for plotting
        if "x_bounds" in kwargs:
            lower, upper = kwargs.pop("x_bounds")
        elif fun != "ppf":
            lower, upper = self.ppf(0.001), self.ppf(0.999)

        if fun == "ppf":
            lower, upper = 0.001, 0.999

        x_arr = np.linspace(lower, upper, 1000)
        y_arr = [getattr(self, fun)(x) for x in x_arr]
        y_arr = np.array(y_arr)

        if ax is None:
            ax = plt.gca()

        ax.plot(x_arr, y_arr, **kwargs)

        if print_labels == "on":
            ax.set_xlabel(f"{x_argname}")
            ax.set_ylabel(f"{fun}({x_argname})")
        return ax


class _Indexer:
    """Indexer for BaseDistribution, for pandas-like index in loc and iloc property."""

    def __init__(self, ref, method="_loc"):
        self.ref = ref
        self.method = method

    def __call__(self, *args, **kwargs):
        """Error message to tell the user to use [ ] instead of ( )."""
        methodname = self.method[1:]
        clsname = self.ref.__class__.__name__
        raise ValueError(
            f"Error while attempting to index {clsname} probability "
            f"distribution instance via {methodname}: "
            f"Please use square brackets [] for indexing a distribution, i.e., "
            f"mydist.{methodname}[index] or mydist.{methodname}[index1, index2], "
            f"not mydist.{methodname}(index) or mydist.{methodname}(index1, index2)"
        )

    def __getitem__(self, key):
        """Getitem dunder, for use in my_distr.loc[index] and my_distr.iloc[index]."""

        def is_noneslice(obj):
            res = isinstance(obj, slice)
            res = res and obj.start is None and obj.stop is None and obj.step is None
            return res

        ref = self.ref
        indexer = getattr(ref, self.method)

        if isinstance(key, tuple):
            if not len(key) == 2:
                raise ValueError(
                    "there should be one or two keys when calling .loc, "
                    "e.g., mydist[key], or mydist[key1, key2]"
                )
            rows = key[0]
            cols = key[1]
            if is_noneslice(rows) and is_noneslice(cols):
                return ref
            elif is_noneslice(cols):
                return indexer(rowidx=rows, colidx=None)
            elif is_noneslice(rows):
                return indexer(rowidx=None, colidx=cols)
            else:
                return indexer(rowidx=rows, colidx=cols)
        else:
            return indexer(rowidx=key, colidx=None)


class _BaseTFDistribution(BaseDistribution):
    """Adapter for tensorflow-probability distributions."""

    _tags = {
        "python_dependencies": "tensorflow_probability",
        "capabilities:approx": ["energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
    }

    def __init__(self, index=None, columns=None, distr=None):
        self.distr = distr

        super().__init__(index=index, columns=columns)

    def __str__(self):
        return self.to_str()

    def pdf(self, x):
        r"""Probability density function.

        Let :math:`X` be a random variables with the distribution of `self`,
        taking values in `(N, n)` `DataFrame`-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`p_{X_{ij}}`, denote the marginal pdf of :math:`X` at the
        :math:`(i,j)`-th entry.

        The output of this method, for input `x` representing :math:`x`,
        is a `DataFrame` with same columns and indices as `self`,
        and entries :math:`p_{X_{ij}}(x_{ij})`.

        If `self` has a mixed or discrete distribution, this returns
        the weighted continuous part of `self`'s distribution instead of the pdf,
        i.e., the marginal pdf integrate to the weight of the continuous part.

        Parameters
        ----------
        x : `pandas.DataFrame` or 2D np.ndarray
            representing :math:`x`, as above

        Returns
        -------
        `DataFrame` with same columns and index as `self`
            containing :math:`p_{X_{ij}}(x_{ij})`, as above
        """
        if isinstance(x, pd.DataFrame):
            dist_at_x = self.loc[x.index, x.columns]
            tensor = dist_at_x.distr.prob(x.values)
            return pd.DataFrame(tensor, index=x.index, columns=x.columns)
        else:
            dist_at_x = self
            return dist_at_x.distr.prob(x)

    def log_pdf(self, x):
        r"""Logarithmic probability density function.

        Numerically more stable than calling pdf and then taking logartihms.

        Let :math:`X` be a random variables with the distribution of `self`,
        taking values in `(N, n)` `DataFrame`-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`p_{X_{ij}}`, denote the marginal pdf of :math:`X` at the
        :math:`(i,j)`-th entry.

        The output of this method, for input `x` representing :math:`x`,
        is a `DataFrame` with same columns and indices as `self`,
        and entries :math:`\log p_{X_{ij}}(x_{ij})`.

        If `self` has a mixed or discrete distribution, this returns
        the weighted continuous part of `self`'s distribution instead of the pdf,
        i.e., the marginal pdf integrate to the weight of the continuous part.

        Parameters
        ----------
        x : `pandas.DataFrame` or 2D np.ndarray
            representing :math:`x`, as above

        Returns
        -------
        `DataFrame` with same columns and index as `self`
            containing :math:`\log p_{X_{ij}}(x_{ij})`, as above
        """
        if isinstance(x, pd.DataFrame):
            dist_at_x = self.loc[x.index, x.columns]
            tensor = dist_at_x.distr.log_prob(x.values)
            return pd.DataFrame(tensor, index=x.index, columns=x.columns)
        else:
            dist_at_x = self
            return dist_at_x.distr.log_prob(x)

    def cdf(self, x):
        """Cumulative distribution function."""
        if isinstance(x, pd.DataFrame):
            dist_at_x = self.loc[x.index, x.columns]
            tensor = dist_at_x.distr.cdf(x.values)
            return pd.DataFrame(tensor, index=x.index, columns=x.columns)
        else:
            dist_at_x = self
            return dist_at_x.distr.cdf(x)

    def sample(self, n_samples=None):
        """Sample from the distribution.

        Parameters
        ----------
        n_samples : int, optional, default = None
            number of samples to draw from the distribution

        Returns
        -------
        pd.DataFrame
            samples from the distribution

            * if ``n_samples`` is ``None``:
            returns a sample that contains a single sample from ``self``,
            in ``pd.DataFrame`` mtype format convention, with ``index`` and ``columns``
            as ``self``
            * if n_samples is ``int``:
            returns a ``pd.DataFrame`` that contains ``n_samples`` i.i.d.
            samples from ``self``, in ``pd-multiindex`` mtype format convention,
            with same ``columns`` as ``self``, and row ``MultiIndex`` that is product
            of ``RangeIndex(n_samples)`` and ``self.index``
        """
        if n_samples is None:
            np_spl = self.distr.sample().numpy()
            return pd.DataFrame(np_spl, index=self.index, columns=self.columns)
        else:
            np_spl = self.distr.sample(n_samples).numpy()
            np_spl = np_spl.reshape(-1, np_spl.shape[-1])
            mi = _prod_multiindex(range(n_samples), self.index)
            df_spl = pd.DataFrame(np_spl, index=mi, columns=self.columns)
            return df_spl


def _prod_multiindex(ix1, ix2):
    rows = []

    def add_rows(rows, ix):
        if isinstance(ix, pd.MultiIndex):
            ix = ix.to_frame()
            rows += [ix[col] for col in ix.columns]
        else:
            rows += [ix]
        return rows

    rows = add_rows(rows, ix1)
    rows = add_rows(rows, ix2)
    res = pd.MultiIndex.from_product(rows)
    res.names = [None] * len(res.names)
    return res


def is_scalar_notnone(obj):
    """Check if obj is scalar and not None."""
    return obj is not None and np.isscalar(obj)


def _get_first_argname(fun):
    """Get the name of the first argument of a function as str."""
    from inspect import signature

    return list(signature(fun).parameters.keys())[0]


def _coerce_to_pd_index_or_none(x):
    """Coerce to pd.Index, if not None, else return None."""
    if x is None:
        return None
    if isinstance(x, pd.Index):
        return x
    return pd.Index(x)



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Empirical quantile parametrized distribution."""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd

from skpro.distributions.empirical import Empirical


class QPD_Empirical(Empirical):
    r"""Empirical quantile parametrized distribution.

    This distribution is parameterized by a set of quantile points and quantiles,
    quantiles :math:`q_1, q_2, \dots, q_N`
    at quantile points :math:`p_1, p_2, \dots, p_N`,
    with :math:`0 \le p_1 < p_2 < \dots < p_N \le 1`.

    It represents a distribution with piecewise constant CDF and quantile function,
    the unique distribution satisfying:

    * the support is :math:`[q_1, q_N]`
    * for any quantile point :math:`p \in [p_1, p_N]`, it holds that
      :math:`\mbox{ppf}(p)` = :math:`\mbox{ppf}(p_i)`,
      where :math:`i` is the index minimizing :math:`|p_i - p|`,
      in all cases where this minimizer is unique.

    In vernacular terms, the quantile function agrees with the quantiles prescribed by
    :math:`q_i` at the quantile points :math:`p_i`, and for other quantile points
    agrees with the value at the nearest quantile point.

    In explicit terms, the distribution is an empirical distribution (sum-of-diracs),
    supported at the quantiles :math:`q_1, q_2, \dots, q_N`,
    with weights :math:`w_1, w_2, \dots, w_N`
    such that :math:`w_i = (p_{i+1} - p_{i-1})/2` for :math:`1 = 1, \dots, N`,
    where we define :math:`p_0 = -p_1` and :math:`p_{N+1} = 2 - p_N`.

    Formally, the distribution is parametrized by the quantiles :math:`q_i`
    and the quantile points :math:`p_i`, not by the quantiles and weights :math:`w_i`,
    so it is distinct from the empirical distribution (``skpro`` ``Empirical``),
    as a parameterized distribution,
    by being quantile parameterized and not sample parameterized.

    However, it is equivalent, as an unparameterized distribution,
    to an ``Empirical`` distribution with weights and nodes given as above.

    Parameters
    ----------
    quantiles : pd.DataFrame with pd.MultiIndex
        quantile points
        first (lowest) index must be float, corresponding to quantile
        further indices are sample indices
    time_indep : bool, optional, default=True
        if True, ``sample`` will sample individual instance indices independently
        if False, ``sample`` will sample entire instances from ``spl``
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> import pandas as pd
    >>> from skpro.distributions import QPD_Empirical

    >>> spl_idx = pd.MultiIndex.from_product(
    ...     [[0.2, 0.5, 0.8], [0, 1, 2]], names=["alpha", "sample"]
    ... )
    >>> spl = pd.DataFrame(
    ...     [[0, 1], [2, 3], [4, 5], [1, 2], [4, 5], [7, 8], [2, 3], [6, 7], [10, 11]],
    ...     index=spl_idx,
    ...     columns=["a", "b"],
    ... )
    >>> dist = QPD_Empirical(spl)
    >>> empirical_sample = dist.sample(3)
    """

    _tags = {
        "capabilities:approx": [],
        "capabilities:exact": ["mean", "var", "energy", "cdf", "ppf"],
        "distr:measuretype": "discrete",
    }

    def __init__(self, quantiles, time_indep=True, index=None, columns=None):
        self.quantiles = quantiles
        self.time_indep = time_indep
        self.index = index
        self.columns = columns

        super().__init__(
            spl=quantiles,
            weights=self._get_empirical_weighted_sample(quantiles),
            time_indep=time_indep,
            index=index,
            columns=columns,
        )

    def _get_empirical_weighted_sample(self, empirical_spl):
        """Compute quantile weights for empirical distribution."""
        alphas = empirical_spl.index.get_level_values(0).unique()
        alpha_sorted = sorted(alphas)

        # obtain alpha weights for empirical distr such that we take the nearest
        # available quantile prob (the cum weights match the chosen quantile prob)
        alpha_np = np.array(alpha_sorted)
        alpha_diff = np.diff(alpha_np)
        alpha_diff2 = np.repeat(alpha_diff, 2) / 2
        weight_double = np.concatenate([[alpha_np[0]], alpha_diff2, [1 - alpha_np[-1]]])
        weight_double2 = weight_double.reshape(-1, 2)
        weights = weight_double2.sum(axis=1)

        # obtain weights per empirical sample
        empirical_spl_weights = pd.Series(index=empirical_spl.index)
        for i, a in enumerate(alpha_sorted):
            empirical_spl_weights.loc[a] = weights[i]

        return empirical_spl_weights

    def _iloc(self, rowidx=None, colidx=None):
        index = self.index
        columns = self.columns

        spl_subset = self.quantiles

        if rowidx is not None:
            rowidx_loc = index[rowidx]
            # subset multiindex to rowidx by last level
            spl_subset = self.spl.loc[(slice(None), rowidx_loc), :]
            subs_rowidx = index[rowidx]
        else:
            subs_rowidx = index

        if colidx is not None:
            spl_subset = spl_subset.iloc[:, colidx]
            subs_colidx = columns[colidx]
        else:
            subs_colidx = columns

        return QPD_Empirical(
            spl_subset,
            time_indep=self.time_indep,
            index=subs_rowidx,
            columns=subs_colidx,
        )

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # params1 is a DataFrame with simple row multiindex
        spl_idx = pd.MultiIndex.from_product(
            [[0.2, 0.5, 0.8], [0, 1, 2]], names=["alpha", "sample"]
        )
        spl = pd.DataFrame(
            [[0, 1], [2, 3], [4, 5], [1, 2], [4, 5], [7, 8], [2, 3], [6, 7], [10, 11]],
            index=spl_idx,
            columns=["a", "b"],
        )
        params1 = {
            "quantiles": spl,
            "time_indep": True,
            "index": pd.RangeIndex(3),
            "columns": pd.Index(["a", "b"]),
        }

        params2 = {
            "quantiles": spl,
            "time_indep": False,
            "index": pd.RangeIndex(3),
            "columns": pd.Index(["a", "b"]),
        }
        return [params1, params2]



"""Module containing adapters other framework packages covering multiple tasks."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["_SksurvAdapter"]

from skpro.survival.adapters.sksurv import _SksurvAdapter



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Adapter to ngboost probabilistic and survival regressors."""

__author__ = ["ShreeshaM07"]


class NGBoostAdapter:
    """Adapter to interconvert NGBoost and skpro BaseDistributions.

    NGBoostAdapter is a Mixin class adapter for the NGBoostRegressor
    and NGBoostSurvival classes to convert the distributions
    from ngboost to skpro and vice versa.

    It uses a boolean survival parameter to distinguish
    whether it is being called from the NGBoostRegressor
    or from NGBoostSurvival.
    """

    def _dist_to_ngboost_instance(self, dist, survival=False):
        """Convert string to NGBoost object.

        Parameters
        ----------
        dist : string
            the input string for the type of Distribution.
            It then creates an object of that particular NGBoost Distribution.
        survival : boolean, default = False
            It denotes whether it is called from the NGBoostSurvival
            or from NGBoostRegressor.

        Returns
        -------
        NGBoost Distribution object.
        """
        from ngboost.distns import Exponential, Laplace, LogNormal, Normal, Poisson, T

        ngboost_dists = {
            "Normal": Normal,
            "Laplace": Laplace,
            "TDistribution": T,
            "Poisson": Poisson,
            "LogNormal": LogNormal,
            "Exponential": Exponential,
        }
        # default Normal distribution
        dist_ngboost = Normal
        # default LogNormal distribution if Survival prediction
        if survival is True:
            dist_ngboost = LogNormal
        # replace default with other distribution if present
        if dist in ngboost_dists:
            dist_ngboost = ngboost_dists[dist]

        return dist_ngboost

    def _ngb_skpro_dist_params(
        self,
        pred_dist,
        index,
        columns,
        **kwargs,
    ):
        import numpy as np

        # The returned values of the Distributions from NGBoost
        # are different. So based on that they are split into these
        # categories of loc,scale,mu and s.
        # Distribution type | Parameters
        # ------------------|-----------
        # Normal            | loc = mean, scale = standard deviation
        # TDistribution     | loc = mean, scale = standard deviation
        # Poisson           | mu = mean
        # LogNormal         | s = standard deviation, scale = exp(mean)
        #                   |     (see scipy.stats.lognorm)
        # Laplace           | loc = mean, scale = scale parameter
        # Exponential       | scale = 1/rate
        # Normal, Laplace, TDistribution and Poisson have not yet
        # been implemented for Survival analysis.

        dist_params = {
            "Normal": ["loc", "scale"],
            "Laplace": ["loc", "scale"],
            "TDistribution": ["loc", "scale"],
            "Poisson": ["mu"],
            "LogNormal": ["scale", "s"],
            "Exponential": ["scale"],
        }

        skpro_params = {
            "Normal": ["mu", "sigma"],
            "Laplace": ["mu", "scale"],
            "TDistribution": ["mu", "sigma"],
            "Poisson": ["mu"],
            "LogNormal": ["mu", "sigma"],
            "Exponential": ["rate"],
        }

        if self.dist in dist_params and self.dist in skpro_params:
            ngboost_params = dist_params[self.dist]
            skp_params = skpro_params[self.dist]
            for ngboost_param, skp_param in zip(ngboost_params, skp_params):
                kwargs[skp_param] = pred_dist.params[ngboost_param]
                if self.dist == "LogNormal" and ngboost_param == "scale":
                    kwargs[skp_param] = np.log(pred_dist.params[ngboost_param])
                if self.dist == "Exponential" and ngboost_param == "scale":
                    kwargs[skp_param] = 1 / pred_dist.params[ngboost_param]

                kwargs[skp_param] = self._check_y(y=kwargs[skp_param])
                # returns a tuple so taking only first index of the tuple
                kwargs[skp_param] = kwargs[skp_param][0]
            kwargs["index"] = index
            kwargs["columns"] = columns

        return kwargs

    def _ngb_dist_to_skpro(self, **kwargs):
        """Convert NGBoost distribution object to skpro BaseDistribution object.

        Parameters
        ----------
        pred_mean, pred_std and index and columns.

        Returns
        -------
        skpro_dist (skpro.distributions.BaseDistribution):
        Converted skpro distribution object.
        """
        from skpro.distributions.exponential import Exponential
        from skpro.distributions.laplace import Laplace
        from skpro.distributions.lognormal import LogNormal
        from skpro.distributions.normal import Normal
        from skpro.distributions.poisson import Poisson
        from skpro.distributions.t import TDistribution

        ngboost_dists = {
            "Normal": Normal,
            "Laplace": Laplace,
            "TDistribution": TDistribution,
            "Poisson": Poisson,
            "LogNormal": LogNormal,
            "Exponential": Exponential,
        }

        skpro_dist = None

        if self.dist in ngboost_dists:
            skpro_dist = ngboost_dists[self.dist](**kwargs)

        return skpro_dist



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Composition for mean/scale family of distributions."""

__author__ = ["fkiraly"]

import numpy as np

from skpro.distributions.base import BaseDistribution


class MeanScale(BaseDistribution):
    r"""Composition for mean/scale family of distributions.

    This distribution parameterizes the natural mean/scale family of distirbutions
    generated by the component distribution ``d``,
    parameterized a mean offset :math:`\mu`, represented by ``mu``, and
    scale factor :math:`s`, represented by ``scale``

    That is, if :math:`X` is distributed according to ``d``, then
    this distribution represents for parameters ``d``, ``mu``, ``scale``,
    the distribution of the random variable :math:`\mu + s \cdot X`.

    For instance, the cdf of this distribution is

    .. math:: F(x) = F_X\left(\frac{1}{s}(x - \mu)\right)

    Parameters
    ----------
    d : BaseDistribution, skpro distribution instance
        component distribution
    mu : float or array of float (1D or 2D)
        mean of the normal distribution
    sigma : float or array of float (1D or 2D), must be positive
        standard deviation of the normal distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.normal import Normal
    >>> from skpro.distributions.meanscale import MeanScale
    >>>
    >>> n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=2)
    >>> d = MeanScale(d=n, mu=2, sigma=3)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["mean", "var", "energy", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "broadcast_init": "on",
        "broadcast_params": ["mu", "sigma"],
    }

    def __init__(self, d, mu=0, sigma=1, index=None, columns=None):
        self.d = d
        self.mu = mu
        self.sigma = sigma

        if d.index is not None and index is None:
            index = d.index
        if d.columns is not None and columns is None:
            columns = d.columns

        super().__init__(index=index, columns=columns)

    # def _iloc(self, rowidx=None, colidx=None):
    #     dist_subset = self.d.iloc[rowidx, colidx]

    #     return MeanScale(d=dist_subset, mu=self.mu, sigma=self.sigma)

    # def _iat(self, rowidx=None, colidx=None):
    #     dist_subset = self.d.iat[rowidx, colidx]
    #     return MeanScale(d=dist_subset, mu=self.mu, sigma=self.sigma)

    def _subset_params(self, rowidx, colidx, coerce_scalar=False):
        """Subset distribution parameters to given rows and columns.

        Parameters
        ----------
        rowidx : None, numpy index/slice coercible, or int
            Rows to subset to. If None, no subsetting is done.
        colidx : None, numpy index/slice coercible, or int
            Columns to subset to. If None, no subsetting is done.
        coerce_scalar : bool, optional, default=False
            If True, and the subsetted parameter is a scalar, coerce it to a scalar.

        Returns
        -------
        dict
            Dictionary with subsetted distribution parameters.
            Keys are parameter names of ``self``, values are the subsetted parameters.
        """
        params = self.get_params()
        subset_params = ["mu", "sigma"]

        subset_param_dict = {}
        for param in subset_params:
            val = params[param]
            subset_param_dict[param] = self._subset_param(
                val=val,
                rowidx=rowidx,
                colidx=colidx,
                coerce_scalar=coerce_scalar,
            )
        if coerce_scalar:
            subset_param_dict["d"] = self.d.iat[rowidx, colidx]
        else:
            subset_param_dict["d"] = self.d.iloc[rowidx, colidx]
        return subset_param_dict

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        return self._bc_params["mu"] + self.d.mean()

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        return self._bc_params["sigma"] ** 2 * self.d.var()

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["sigma"]

        x_inner = (x - mu) / scale
        pdf_arr = self.d.pdf(x_inner) / scale
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["sigma"]

        x_inner = (x - mu) / scale
        lpdf_arr = self.d.log_pdf(x_inner) - np.log(scale)
        return lpdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["sigma"]

        x_inner = (x - mu) / scale
        cdf_arr = self.d.cdf(x_inner)
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["sigma"]

        icdf_arr = mu + scale * self.d.ppf(p)
        return icdf_arr

        # def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        # scale = self._bc_params["sigma"]

        # en_arr = scale * self.d.energy()
        # return en_arr

        # def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        # mu = self._bc_params["mu"]
        # scale = self._bc_params["sigma"]

        # x_offset = (x - mu) / scale
        # en_arr = scale * self.d.energy(x_offset)
        # return en_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        from skpro.distributions.normal import Normal

        d = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=2)
        dsc = Normal(0, 1)

        # array case examples
        params0 = {"d": d}
        params1 = {"d": d, "mu": 1}
        params2 = {"d": d, "sigma": 2}
        params3 = {"d": d, "mu": 2, "sigma": 3}
        params4 = {"d": d, "mu": [[0, 1], [2, 3], [4, 5]], "sigma": 2}
        # params5 = {
        #     "d": dsc,
        #     "mu": 0,
        #     "sigma": 1,
        #     "index": pd.Index([1, 2, 5]),
        #     "columns": pd.Index(["a", "b"]),
        # }

        # scalar case examples
        params6 = {"d": dsc}
        params7 = {"d": dsc, "mu": 1, "sigma": 2}
        return [
            params0,
            params1,
            params2,
            params3,
            params4,
            # params5,
            params6,
            params7,
        ]



"""Tests for evaluate utility."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# based on the sktime tests of the same name

__author__ = ["fkiraly"]


import numpy as np
import pandas as pd
import pytest
from skbase.utils.dependencies import _check_soft_dependencies
from sklearn.datasets import load_diabetes
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import KFold, ShuffleSplit

from skpro.benchmarking.evaluate import evaluate
from skpro.metrics import CRPS, EmpiricalCoverage, LogLoss, PinballLoss
from skpro.regression.residual import ResidualDouble
from skpro.tests.test_switch import run_test_module_changed


def _check_evaluate_output(out, cv, y, scoring):
    assert isinstance(out, pd.DataFrame)

    if hasattr(scoring, "get_tag"):
        scitype = scoring.get_tag("scitype:y_pred", raise_error=False)
    else:
        scitype = None

    # Check column names.
    assert set(out.columns) == {
        "fit_time",
        "len_y_train",
        f"{scitype}_time",
        f"test_{scoring.name}",
    }

    # Check number of rows against number of splits.
    n_splits = cv.get_n_splits(y)
    assert out.shape[0] == n_splits

    # Check that all timings are positive.
    assert np.all(out.filter(like="_time") >= 0)

    # Check training set sizes
    assert np.all(out["len_y_train"] > 0)


def _get_pred_method(scoring):
    """Get the prediction method for a given scoring function."""
    pred_type = {
        "pred_quantiles": "predict_quantiles",
        "pred_interval": "predict_interval",
        "pred_proba": "predict_proba",
        None: "predict",
    }

    if hasattr(scoring, "get_tag"):
        scitype = scoring.get_tag("scitype:y_pred", raise_error=False)
    else:
        scitype = None

    return pred_type[scitype]


CVs = [
    KFold(n_splits=3),
    ShuffleSplit(n_splits=3, test_size=0.5, random_state=42),
]

METRICS = [CRPS, EmpiricalCoverage, LogLoss, PinballLoss]


@pytest.mark.skipif(
    not run_test_module_changed(["skpro.benchmarking", "skpro.metrics"]),
    reason="Test only if skpro.benchmarking or skpro.metrics has been changed",
)
@pytest.mark.parametrize("cv", CVs)
@pytest.mark.parametrize("scoring", METRICS)
@pytest.mark.parametrize("backend", [None, "dask", "loky", "threading"])
def test_evaluate_common_configs(cv, scoring, backend):
    """Test evaluate common configs."""
    # skip test for dask backend if dask is not installed
    if backend == "dask" and not _check_soft_dependencies("dask", severity="none"):
        return None

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    y = pd.DataFrame(y)
    estimator = ResidualDouble(LinearRegression(), min_scale=1)

    scoring = scoring()

    out = evaluate(
        estimator=estimator,
        X=X,
        y=y,
        cv=cv,
        scoring=scoring,
        backend=backend,
    )
    _check_evaluate_output(out, cv, y, scoring)

    # check scoring
    actual = out.loc[:, f"test_{scoring.name}"]

    n_splits = cv.get_n_splits(X)
    expected = np.empty(n_splits)

    for i, (train, test) in enumerate(cv.split(y)):
        X_train, y_train = X.iloc[train], y.iloc[train]
        X_test, y_test = X.iloc[test], y.iloc[test]
        est = estimator.clone()
        est.fit(X_train, y_train)

        pred_method = _get_pred_method(scoring)
        y_pred = getattr(est, pred_method)(X_test)
        expected[i] = scoring(y_test, y_pred, y_train=y_train)

    np.testing.assert_array_equal(actual, expected)



"""Test dummy for testing config skips."""


from skpro.regression.base import BaseProbaRegressor  # noqa: E402


class DummySkipped(BaseProbaRegressor):
    """Dummy regressor to test exclusion."""

    pass



"""Registry of mtypes for Table scitype.

See datatypes._registry for API.
"""

import pandas as pd

__all__ = [
    "MTYPE_REGISTER_TABLE",
    "MTYPE_LIST_TABLE",
]


MTYPE_REGISTER_TABLE = [
    ("pd_DataFrame_Table", "Table", "pd.DataFrame representation of a data table"),
    ("numpy1D", "Table", "1D np.narray representation of a univariate table"),
    ("numpy2D", "Table", "2D np.narray representation of a univariate table"),
    ("pd_Series_Table", "Table", "pd.Series representation of a data table"),
    ("list_of_dict", "Table", "list of dictionaries with primitive entries"),
    ("polars_eager_table", "Table", "polars.DataFrame representation of a data table"),
    ("polars_lazy_table", "Table", "polars.LazyFrame representation of a data table"),
]

MTYPE_LIST_TABLE = pd.DataFrame(MTYPE_REGISTER_TABLE)[0].values



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Example fixtures for mtypes/scitypes.

Exports
-------
get_examples(mtype: str, as_scitype: str, return_lossy=False)
    retrieves examples for mtype/scitype, and/or whether it is lossy

examples[i] are assumed "content-wise the same" for the same as_scitype, i
    only in different machine representations

the representation is considered "lossy" if the representation is incomplete
    e.g., metadata such as column names are missing
"""

from copy import deepcopy
from functools import lru_cache

from skpro.datatypes._registry import mtype_to_scitype

__author__ = ["fkiraly"]

__all__ = [
    "get_examples",
]


@lru_cache(maxsize=1)
def generate_example_dicts(soft_deps="present"):
    """Generate example dicts using lookup."""
    from skbase.utils.dependencies import _check_estimator_deps

    from skpro.datatypes._base import BaseExample
    from skpro.utils.retrieval import _all_classes

    classes = _all_classes("skpro.datatypes")
    classes = [x[1] for x in classes]
    classes = [x for x in classes if issubclass(x, BaseExample)]
    classes = [x for x in classes if not x.__name__.startswith("Base")]

    # subset only to data types with soft dependencies present
    if soft_deps == "present":
        classes = [x for x in classes if _check_estimator_deps(x, severity="none")]

    example_dict = dict()
    example_dict_lossy = dict()
    example_dict_metadata = dict()
    for cls in classes:
        k = cls()
        key = k._get_key()
        key_meta = (key[1], key[2])
        example_dict[key] = k
        example_dict_lossy[key] = k.get_class_tags().get("lossy", False)
        example_dict_metadata[key_meta] = k.get_class_tags().get("metadata", {})

    return example_dict, example_dict_lossy, example_dict_metadata


def get_examples(
    mtype: str,
    as_scitype: str = None,
    return_lossy: bool = False,
    return_metadata: bool = False,
):
    """Retrieve a dict of examples for mtype `mtype`, scitype `as_scitype`.

    Parameters
    ----------
    mtype: str - name of the mtype for the example, a valid mtype string
        valid mtype strings, with explanation, are in datatypes.MTYPE_REGISTER
    as_scitype : str, optional - name of scitype of the example, a valid scitype string
        valid scitype strings, with explanation, are in datatypes.SCITYPE_REGISTER
        default = inferred from mtype of obj
    return_lossy: bool, optional, default=False
        whether to return second argument
    return_metadata: bool, optional, default=False
        whether to return third argument

    Returns
    -------
    fixtures: dict with integer keys, elements being
        fixture - example for mtype `mtype`, scitype `as_scitype`
        if return_lossy=True, elements are pairs with fixture and
            lossy: bool - whether the example is a lossy representation
        if return_metadata=True, elements are triples with fixture, lossy, and
            metadata: dict - metadata dict that would be returned by check_is_mtype
    """
    # if as_scitype is None, infer from mtype
    if as_scitype is None:
        as_scitype = mtype_to_scitype(mtype)

    example_dict, example_dict_lossy, example_dict_metadata = generate_example_dicts()

    # retrieve all keys that match the query
    exkeys = example_dict.keys()
    keys = [k for k in exkeys if k[0] == mtype and k[1] == as_scitype]

    # retrieve all fixtures that match the key
    fixtures = dict()

    for k in keys:
        if return_lossy:
            fixtures[k[2]] = (example_dict.get(k).build(), example_dict_lossy.get(k))
        elif return_metadata:
            fixtures[k[2]] = (
                example_dict.get(k).build(),
                example_dict_lossy.get(k),
                example_dict_metadata.get((k[1], k[2])),
            )
        else:
            fixtures[k[2]] = example_dict.get(k).build()

    # deepcopy to avoid side effects
    return deepcopy(fixtures)



"""Testing utility to compare equality in value for nested objects.

Objects compared can have one of the following valid types:
    types compatible with != comparison
    pd.Series, pd.DataFrame, np.ndarray
    lists, tuples, or dicts of a valid type (recursive)
    polars.DataFrame, polars.LazyFrame
"""
from skbase.utils.deep_equals._common import _make_ret
from skbase.utils.deep_equals._deep_equals import deep_equals as _deep_equals
from skbase.utils.dependencies import _check_soft_dependencies

__author__ = ["fkiraly"]
__all__ = ["deep_equals"]


def deep_equals(x, y, return_msg=False, plugins=None):
    """Test two objects for equality in value.

    Correct if x/y are one of the following valid types:
        types compatible with != comparison
        pd.Series, pd.DataFrame, np.ndarray
        lists, tuples, or dicts of a valid type (recursive)

    Important note:
        this function will return "not equal" if types of x,y are different
        for instant, bool and numpy.bool are *not* considered equal

    Parameters
    ----------
    x : object
    y : object
    return_msg : bool, optional, default=False
        whether to return informative message about what is not equal
    plugins : list, optional, default=None
        optional additional deep_equals plugins to use
        will be appended to the default plugins from ``deep_equals_custom``
        see ``deep_equals_custom`` for details of signature of plugins

    Returns
    -------
    is_equal: bool - True if x and y are equal in value
        x and y do not need to be equal in reference
    msg : str, only returned if return_msg = True
        indication of what is the reason for not being equal
            concatenation of the following strings:
            .type - type is not equal
            .class - both objects are classes but not equal
            .len - length is not equal
            .value - value is not equal
            .keys - if dict, keys of dict are not equal
                    if class/object, names of attributes and methods are not equal
            .dtype - dtype of pandas or numpy object is not equal
            .index - index of pandas object is not equal
            .series_equals, .df_equals, .index_equals - .equals of pd returns False
            [i] - if tuple/list: i-th element not equal
            [key] - if dict: value at key is not equal
            [colname] - if pandas.DataFrame: column with name colname is not equal
            != - call to generic != returns False
            .polars_equals - .equals of polars returns False
    """
    # call deep_equals_custom with default plugins
    plugins_default = [
        _polars_equals_plugin,
    ]

    if plugins is not None:
        plugins_inner = plugins_default + plugins
    else:
        plugins_inner = plugins_default

    res = _deep_equals(x, y, return_msg=return_msg, plugins=plugins_inner)
    return res


def _polars_equals_plugin(x, y, return_msg=False):
    polars_available = _check_soft_dependencies("polars", severity="none")

    if not polars_available:
        return None

    import polars as pl

    if not isinstance(x, (pl.DataFrame, pl.LazyFrame)):
        return None

    ret = _make_ret(return_msg)

    # compare pl.DataFrame
    if isinstance(x, pl.DataFrame):
        return ret(x.equals(y), ".polars_equals")

    # compare pl.LazyFrame
    if isinstance(x, pl.LazyFrame):
        return ret(x.collect().equals(y.collect()), ".polars_equals")

    return None



#!/usr/bin/env python3 -u
# License: BSD 3 clause
"""Utility methods to print system info for debugging.

adapted from
:func: `sklearn.show_versions`
"""

__author__ = ["mloning", "fkiraly"]
__all__ = ["show_versions"]

import importlib
import platform
import sys


def _get_sys_info():
    """System information.

    Return
    ------
    sys_info : dict
        system and Python version information
    """
    python = sys.version.replace("\n", " ")

    blob = [
        ("python", python),
        ("executable", sys.executable),
        ("machine", platform.platform()),
    ]

    return dict(blob)


# dependencies to print versions of, by default
DEFAULT_DEPS_TO_SHOW = [
    "pip",
    "sktime",
    "sklearn",
    "skbase",
    "numpy",
    "scipy",
    "pandas",
    "matplotlib",
    "joblib",
    "numba",
    "statsmodels",
    "pmdarima",
    "statsforecast",
    "tsfresh",
    "tslearn",
    "torch",
    "tensorflow",
    "tensorflow_probability",
]


def _get_deps_info(deps=None):
    """Overview of the installed version of main dependencies.

    Parameters
    ----------
    deps : optional, list of strings with import names
        if None, behaves as deps = ["sktime"]

    Returns
    -------
    deps_info: dict
        version information on libraries in `deps`
        keys are import names, values are PEP 440 version strings
        of the import as present in the current python environment
    """
    if deps is None:
        deps = ["sktime"]

    def get_version(module):
        return getattr(module, "__version__", None)

    deps_info = {}

    for modname in deps:
        try:
            if modname in sys.modules:
                mod = sys.modules[modname]
            else:
                mod = importlib.import_module(modname)
        except ImportError:
            deps_info[modname] = None
        else:
            ver = get_version(mod)
            deps_info[modname] = ver

    return deps_info


def show_versions():
    """Print python version, OS version, sktime version, selected dependency versions.

    Pretty prints:

    * python version of environment
    * python executable location
    * OS version
    * list of import name and version number for selected python dependencies

    Developer note:
    Python version/executable and OS version are from `_get_sys_info`
    Package versions are retrieved by `_get_deps_info`
    Selected dependencies are as in the DEFAULT_DEPS_TO_SHOW variable
    """
    sys_info = _get_sys_info()
    deps_info = _get_deps_info(deps=DEFAULT_DEPS_TO_SHOW)

    print("\nSystem:")  # noqa: T001, T201
    for k, stat in sys_info.items():
        print(f"{k:>10}: {stat}")  # noqa: T001, T201

    print("\nPython dependencies:")  # noqa: T001, T201
    for k, stat in deps_info.items():
        print(f"{k:>13}: {stat}")  # noqa: T001, T201



"""Base module with base classes BaseObject, BaseEstimator."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["BaseEstimator", "BaseMetaEstimator", "BaseObject"]

from skpro.base._base import BaseEstimator, BaseMetaEstimator, BaseObject



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Utility for benchmark evaluation of probabilistic regression models."""
# based on the sktime utility of the same name

__author__ = ["fkiraly", "hazrulakmal"]
__all__ = ["evaluate"]

import time
import warnings

import numpy as np
import pandas as pd
from skbase.utils.dependencies import _check_soft_dependencies

from skpro.utils.parallel import parallelize


def _split(X, y, C, train, test):
    results = dict()
    # split data according to cv
    results["X_train"] = X.iloc[train]
    results["X_test"] = X.iloc[test]

    if y is not None:
        results["y_train"] = y.iloc[train]
        results["y_test"] = y.iloc[test]

    if C is not None:
        results["C_train"] = C.iloc[train]
        results["C_test"] = C.iloc[test]

    return results


def evaluate(
    estimator,
    cv,
    X,
    y,
    scoring=None,
    return_data=False,
    error_score=np.nan,
    backend=None,
    compute=True,
    backend_params=None,
    C=None,
    **kwargs,
):
    r"""Evaluate estimator using re-sample folds.

    All-in-one statistical performance benchmarking utility for estimators
    which runs a simple backtest experiment and returns a summary pd.DataFrame.

    The experiment run is the following:

    Denote by :math:`X_{train, 1}, X_{test, 1}, \dots, X_{train, K}, X_{test, K}`
    the train/test folds produced by the generator ``cv.split(X)``
    Denote by :math:`y_{train, 1}, y_{test, 1}, \dots, y_{train, K}, y_{test, K}`
    the train/test folds produced by the generator ``cv.split(y)``.

    0. For ``i = 1`` to ``cv.get_n_folds(X)`` do:
    1. ``fit`` the ``estimator`` to :math:`X_{train, 1}`, :math:`y_{train, 1}`
    2. ``y_pred = estimator.predict``
      (or ``predict_proba`` or ``predict_quantiles``, depending on ``scoring``)
      with exogeneous data :math:`X_{test, i}`
    3. Compute ``scoring`` on ``y_pred``versus :math:`y_{test, 1}`.

    Results returned in this function's return are:
    * results of ``scoring`` calculations, from 3,  in the `i`-th loop
    * runtimes for fitting and/or predicting, from 1, 2 in the `i`-th loop
    * :math:`y_{train, i}`, :math:`y_{test, i}`, ``y_pred`` (optional)

    A distributed and-or parallel back-end can be chosen via the ``backend`` parameter.

    Parameters
    ----------
    estimator : skpro BaseProbaRegressor descendant (concrete estimator)
        skpro estimator to benchmark
    cv : sklearn splitter
        determines split of ``X`` and ``y`` into test and train folds
    X : pandas DataFrame
        Feature instances to use in evaluation experiment
    y : pd.DataFrame, must be same length as X
        Labels to used in the evaluation experiment
    scoring : subclass of skpro.performance_metrics.BaseMetric or list of same,
        default=None. Used to get a score function that takes y_pred and y_test
        arguments and accept y_train as keyword argument.
        If None, then uses scoring = MeanAbsolutePercentageError(symmetric=True).
    return_data : bool, default=False
        Returns three additional columns in the DataFrame, by default False.
        The cells of the columns contain each a pd.Series for y_train,
        y_pred, y_test.
    error_score : "raise" or numeric, default=np.nan
        Value to assign to the score if an exception occurs in estimator fitting. If set
        to "raise", the exception is raised. If a numeric value is given,
        FitFailedWarning is raised.

    backend : {"dask", "loky", "multiprocessing", "threading"}, by default None.
        Runs parallel evaluate if specified and `strategy` is set as "refit".

        - "None": executes loop sequentally, simple list comprehension
        - "loky", "multiprocessing" and "threading": uses ``joblib.Parallel`` loops
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``
        - "dask": uses ``dask``, requires ``dask`` package in environment
        - "dask_lazy": same as "dask",
          but changes the return to (lazy) ``dask.dataframe.DataFrame``.

        Recommendation: Use "dask" or "loky" for parallel evaluate.
        "threading" is unlikely to see speed ups due to the GIL and the serialization
        backend (``cloudpickle``) for "dask" and "loky" is generally more robust
        than the standard ``pickle`` library used in "multiprocessing".

    compute : bool, default=True
        If backend="dask", whether returned DataFrame is computed.
        If set to True, returns `pd.DataFrame`, otherwise `dask.dataframe.DataFrame`.

    backend_params : dict, optional
        additional parameters passed to the backend as config.
        Directly passed to ``utils.parallel.parallelize``.
        Valid keys depend on the value of ``backend``:

        - "None": no additional parameters, ``backend_params`` is ignored
        - "loky", "multiprocessing" and "threading": default ``joblib`` backends
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          with the exception of ``backend`` which is directly controlled by ``backend``.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``.
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          ``backend`` must be passed as a key of ``backend_params`` in this case.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "dask": any valid keys for ``dask.compute`` can be passed,
          e.g., ``scheduler``

    C : pd.DataFrame, optional (default=None)
        censoring information to use in the evaluation experiment,
        should have same column name as y, same length as X and y
        should have entries 0 and 1 (float or int)
        0 = uncensored, 1 = (right) censored
        if None, all observations are assumed to be uncensored
        Can be passed to any probabilistic regressor,
        but is ignored if capability:survival tag is False.

    Returns
    -------
    results : pd.DataFrame or dask.dataframe.DataFrame
        DataFrame that contains several columns with information regarding each
        refit/update and prediction of the estimator.
        Row index is splitter index of train/test fold in ``cv``.
        Entries in the i-th row are for the i-th train/test split in ``cv``.
        Columns are as follows:
        - test_{scoring.name}: (float) Model performance score.
          If ``scoring`` is a list,
          then there is a column withname ``test_{scoring.name}`` for each scorer.
        - fit_time: (float) Time in sec for ``fit`` on train fold.
        - pred_time: (float) Time in sec to ``predict`` from fitted estimator.
        - pred_[method]_time: (float)
          Time in sec to run ``predict_[method]`` from fitted estimator.
        - len_y_train: (int) length of y_train.
        - y_train: (pd.Series) only present if see ``return_data=True``
          train fold of the i-th split in ``cv``, used to fit the estimator.
        - y_pred: (pd.Series) present if see ``return_data=True``
          predictions from fitted estimator for the i-th test fold indices of ``cv``.
        - y_test: (pd.Series) present if see ``return_data=True``
          testing fold of the i-th split in ``cv``, used to compute the metric.

    Examples
    --------
    >>> import pandas as pd
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.model_selection import KFold
    >>>
    >>> from skpro.benchmarking.evaluate import evaluate
    >>> from skpro.metrics import CRPS
    >>> from skpro.regression.residual import ResidualDouble
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> y = pd.DataFrame(y)  # skpro assumes y is pd.DataFrame
    >>>
    >>> estimator = ResidualDouble(LinearRegression())
    >>> cv = KFold(n_splits=3)
    >>> crps = CRPS()
    >>>
    >>> results = evaluate(estimator=estimator, X=X, y=y, cv=cv, scoring=crps)
    """
    if backend == "dask" and not _check_soft_dependencies("dask", severity="none"):
        raise RuntimeError(
            "running evaluate with backend='dask' requires the dask package installed,"
            "but dask is not present in the python environment"
        )

    # todo: input checks and coercions
    # cv = check_cv(cv, enforce_start_with_window=True)

    scoring = _check_scores(scoring)

    _evaluate_fold_kwargs = {
        "estimator": estimator,
        "scoring": scoring,
        "return_data": return_data,
        "error_score": error_score,
    }

    def gen_X_y_train_test(X, y, C, cv):
        """Generate joint splits of X, y as per cv.

        Yields
        ------
        X_train : i-th train split of y as per cv.
        X_test : i-th test split of y as per cv.
        y_train : i-th train split of y as per cv. None if y was None.
        y_test : i-th test split of y as per cv. None if y was None.
        C_train : i-th train split of C as per cv. None if C was None.
        C_test : i-th test split of C as per cv. None if C was None.
        """
        for train, test in cv.split(X):
            yield _split(X, y, C, train, test)

    # generator for X and y splits to iterate over below
    xy_splits = gen_X_y_train_test(X, y, C, cv)

    if backend == "dask":
        backend_in = "dask_lazy"
    else:
        backend_in = backend

    # dispatch by backend
    results = parallelize(
        fun=_evaluate_fold,
        iter=xy_splits,
        meta=_evaluate_fold_kwargs,
        backend=backend_in,
        backend_params=backend_params,
    )

    # final formatting of dask dataframes
    if backend in ["dask", "dask_lazy"]:
        import dask.dataframe as dd

        metadata = _get_column_order_and_datatype(scoring, return_data)

        results = dd.from_delayed(results, meta=metadata)
        if backend == "dask":
            results = results.compute()
    else:
        results = pd.concat(results)

    # final formatting of results DataFrame
    results = results.reset_index(drop=True)

    return results


def _evaluate_fold(x, meta):
    # unpack args
    X_train = x["X_train"]
    X_test = x["X_test"]
    y_train = x["y_train"]
    y_test = x["y_test"]
    C_train = x.get("C_train", None)
    C_test = x.get("C_test", None)

    estimator = meta["estimator"]
    scoring = meta["scoring"]
    return_data = meta["return_data"]
    error_score = meta["error_score"]

    # set default result values in case estimator fitting fails
    score = error_score
    fit_time = np.nan
    pred_time = np.nan
    y_pred = pd.NA

    # results and cache dictionaries
    temp_result = dict()
    y_preds_cache = dict()

    try:
        # fit/update
        start_fit = time.perf_counter()

        estimator = estimator.clone()
        estimator.fit(X_train, y_train, C=C_train)

        fit_time = time.perf_counter() - start_fit

        pred_type = {
            "pred_quantiles": "predict_quantiles",
            "pred_interval": "predict_interval",
            "pred_proba": "predict_proba",
            None: "predict",
        }
        # predict
        start_pred = time.perf_counter()
        # cache prediction from the first scitype and reuse it to compute other metrics
        for scitype in scoring:
            method = getattr(estimator, pred_type[scitype])
            for metric in scoring.get(scitype):
                pred_args = _get_pred_args_from_metric(scitype, metric)
                if pred_args == {}:
                    time_key = f"{scitype}_time"
                    result_key = f"test_{metric.name}"
                    y_pred_key = f"y_{scitype}"
                else:
                    argval = list(pred_args.values())[0]
                    time_key = f"{scitype}_{argval}_time"
                    result_key = f"test_{metric.name}_{argval}"
                    y_pred_key = f"y_{scitype}_{argval}"

                # make prediction
                if y_pred_key not in y_preds_cache.keys():
                    start_pred = time.perf_counter()
                    y_pred = method(X_test, **pred_args)
                    pred_time = time.perf_counter() - start_pred
                    temp_result[time_key] = [pred_time]
                    y_preds_cache[y_pred_key] = [y_pred]
                else:
                    y_pred = y_preds_cache[y_pred_key][0]

                # score prediction and store score
                score = metric(y_test, y_pred, y_train=y_train, C_true=C_test)
                temp_result[result_key] = [score]

    except Exception as e:
        if error_score == "raise":
            raise e
        else:
            warnings.warn(
                f"""
                In evaluate, fitting of estimator {type(estimator).__name__} failed,
                you can set error_score='raise' in evaluate to see
                the exception message. Fit failed for len(y_train)={len(y_train)}.
                The score will be set to {error_score}.
                Failed estimator with parameters: {estimator}.
                """,
                stacklevel=2,
            )

    # format results data frame and return
    temp_result["fit_time"] = [fit_time]
    temp_result["pred_time"] = [pred_time]
    temp_result["len_y_train"] = [len(y_train)]
    if return_data:
        temp_result["y_train"] = [y_train]
        temp_result["y_test"] = [y_test]
        temp_result.update(y_preds_cache)
    result = pd.DataFrame(temp_result)
    result = result.astype({"len_y_train": int})

    column_order = _get_column_order_and_datatype(scoring, return_data)
    result = result.reindex(columns=column_order.keys())

    return result


def _get_pred_args_from_metric(scitype, metric):
    pred_args = {
        "pred_quantiles": "alpha",
        "pred_interval": "coverage",
    }
    if scitype in pred_args.keys():
        val = getattr(metric, pred_args[scitype], None)
        if val is not None:
            return {pred_args[scitype]: val}
    return {}


def _get_column_order_and_datatype(metric_types, return_data=True):
    """Get the ordered column name and input datatype of results."""
    others_metadata = {"len_y_train": "int"}
    y_metadata = {
        "y_train": "object",
        "y_test": "object",
    }
    fit_metadata, metrics_metadata = {"fit_time": "float"}, {}
    for scitype in metric_types:
        for metric in metric_types.get(scitype):
            pred_args = _get_pred_args_from_metric(scitype, metric)
            if pred_args == {}:
                time_key = f"{scitype}_time"
                result_key = f"test_{metric.name}"
                y_pred_key = f"y_{scitype}"
            else:
                argval = list(pred_args.values())[0]
                time_key = f"{scitype}_{argval}_time"
                result_key = f"test_{metric.name}_{argval}"
                y_pred_key = f"y_{scitype}_{argval}"
            fit_metadata[time_key] = "float"
            metrics_metadata[result_key] = "float"
            if return_data:
                y_metadata[y_pred_key] = "object"
    fit_metadata.update(others_metadata)
    if return_data:
        fit_metadata.update(y_metadata)
    metrics_metadata.update(fit_metadata)
    return metrics_metadata.copy()


def _check_scores(metrics):
    """Validate and coerce to BaseMetric and segregate them based on predict type.

    Parameters
    ----------
    metrics : sktime accepted metrics object or a list of them or None

    Return
    ------
    metrics_type : Dict
        The key is metric types and its value is a list of its corresponding metrics.
    """
    if not isinstance(metrics, list):
        metrics = [metrics]

    metrics_type = {}
    for metric in metrics:
        # collect predict type
        if hasattr(metric, "get_tag"):
            scitype = metric.get_tag(
                "scitype:y_pred", raise_error=False, tag_value_default="pred"
            )
        else:  # If no scitype exists then metric is a point forecast type
            scitype = "pred"
        if scitype not in metrics_type.keys():
            metrics_type[scitype] = [metric]
        else:
            metrics_type[scitype].append(metric)
    return metrics_type



"""Registry lookup methods.

This module exports the following methods for registry lookup:

all_objects(object_types, filter_tags)
    lookup and filtering of objects
"""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# based on the sktime module of same name

__author__ = ["fkiraly"]
# all_objects is based on the sklearn utility all_estimators


from copy import deepcopy
from inspect import isclass
from operator import itemgetter
from pathlib import Path

import pandas as pd
from skbase.lookup import all_objects as _all_objects

from skpro.base import BaseEstimator, BaseObject
from skpro.registry._tags import OBJECT_TAG_REGISTER

VALID_OBJECT_TYPE_STRINGS = {x[1] for x in OBJECT_TAG_REGISTER}


def all_objects(
    object_types=None,
    filter_tags=None,
    exclude_objects=None,
    return_names=True,
    as_dataframe=False,
    return_tags=None,
    suppress_import_stdout=True,
):
    """Get a list of all objects from skpro.

    This function crawls the module and gets all classes that inherit
    from skpro's and sklearn's base classes.

    Not included are: the base classes themselves, classes defined in test
    modules.

    Parameters
    ----------
    object_types: str, list of str, optional (default=None)
        Which kind of objects should be returned.
        if None, no filter is applied and all objects are returned.
        if str or list of str, strings define scitypes specified in search
        only objects that are of (at least) one of the scitypes are returned
        possible str values are entries of registry.BASE_CLASS_REGISTER (first col)
        for instance 'regrssor_proba', 'distribution, 'metric'

    return_names: bool, optional (default=True)

        if True, estimator class name is included in the ``all_objects``
        return in the order: name, estimator class, optional tags, either as
        a tuple or as pandas.DataFrame columns

        if False, estimator class name is removed from the ``all_objects`` return.

    filter_tags: dict of (str or list of str), optional (default=None)
        For a list of valid tag strings, use the registry.all_tags utility.

        ``filter_tags`` subsets the returned estimators as follows:

        * each key/value pair is statement in "and"/conjunction
        * key is tag name to sub-set on
        * value str or list of string are tag values
        * condition is "key must be equal to value, or in set(value)"

    exclude_estimators: str, list of str, optional (default=None)
        Names of estimators to exclude.

    as_dataframe: bool, optional (default=False)

        True: ``all_objects`` will return a pandas.DataFrame with named
        columns for all of the attributes being returned.

        False: ``all_objects`` will return a list (either a list of
        estimators or a list of tuples, see Returns)

    return_tags: str or list of str, optional (default=None)
        Names of tags to fetch and return each estimator's value of.
        For a list of valid tag strings, use the registry.all_tags utility.
        if str or list of str,
        the tag values named in return_tags will be fetched for each
        estimator and will be appended as either columns or tuple entries.

    suppress_import_stdout : bool, optional. Default=True
        whether to suppress stdout printout upon import.

    Returns
    -------
    all_objects will return one of the following:
        1. list of objects, if return_names=False, and return_tags is None
        2. list of tuples (optional object name, class, ~optional object
          tags), if return_names=True or return_tags is not None.
        3. pandas.DataFrame if as_dataframe = True
        if list of objects:
            entries are objects matching the query,
            in alphabetical order of object name
        if list of tuples:
            list of (optional object name, object, optional object
            tags) matching the query, in alphabetical order of object name,
            where
            ``name`` is the object name as string, and is an
                optional return
            ``object`` is the actual object
            ``tags`` are the object's values for each tag in return_tags
                and is an optional return.
        if dataframe:
            all_objects will return a pandas.DataFrame.
            column names represent the attributes contained in each column.
            "objects" will be the name of the column of objects, "names"
            will be the name of the column of object class names and the string(s)
            passed in return_tags will serve as column names for all columns of
            tags that were optionally requested.

    Examples
    --------
    >>> from skpro.registry import all_objects
    >>> # return a complete list of objects as pd.Dataframe
    >>> all_objects(as_dataframe=True)  # doctest: +SKIP
    >>> # return all probabilistic regressors by filtering for object type
    >>> all_objects("regressor_proba", as_dataframe=True)  # doctest: +SKIP
    >>> # return all regressors which handle missing data in the input by tag filtering
    >>> all_objects(
    ...     "regressor_proba",
    ...     filter_tags={"capability:missing": True},
    ...     as_dataframe=True
    ... )  # doctest: +SKIP

    References
    ----------
    Adapted version of sktime's ``all_estimators``,
    which is an evolution of scikit-learn's ``all_estimators``
    """
    MODULES_TO_IGNORE = (
        "tests",
        "setup",
        "contrib",
        "utils",
        "all",
    )

    result = []
    ROOT = str(Path(__file__).parent.parent)  # skpro package root directory

    def _coerce_to_str(obj):
        if isinstance(obj, (list, tuple)):
            return [_coerce_to_str(o) for o in obj]
        if isclass(obj):
            obj = obj.get_tag("object_type")
        return obj

    def _coerce_to_list_of_str(obj):
        obj = _coerce_to_str(obj)
        if isinstance(obj, str):
            return [obj]
        return obj

    if object_types is not None:
        object_types = _coerce_to_list_of_str(object_types)
        object_types = list(set(object_types))

    if object_types is not None:
        if filter_tags is None:
            filter_tags = {}
        elif isinstance(filter_tags, str):
            filter_tags = {filter_tags: True}
        else:
            filter_tags = filter_tags.copy()

        if "object_type" in filter_tags:
            obj_field = filter_tags["object_type"]
            obj_field = _coerce_to_list_of_str(obj_field)
            obj_field = obj_field + object_types
        else:
            obj_field = object_types

        filter_tags["object_type"] = obj_field

    result = _all_objects(
        object_types=[BaseObject, BaseEstimator],
        filter_tags=filter_tags,
        exclude_objects=exclude_objects,
        return_names=return_names,
        as_dataframe=as_dataframe,
        return_tags=return_tags,
        suppress_import_stdout=suppress_import_stdout,
        package_name="skpro",
        path=ROOT,
        modules_to_ignore=MODULES_TO_IGNORE,
    )

    return result


def _check_list_of_str_or_error(arg_to_check, arg_name):
    """Check that certain arguments are str or list of str.

    Parameters
    ----------
    arg_to_check: argument we are testing the type of
    arg_name: str,
        name of the argument we are testing, will be added to the error if
        ``arg_to_check`` is not a str or a list of str

    Returns
    -------
    arg_to_check: list of str,
        if arg_to_check was originally a str it converts it into a list of str
        so that it can be iterated over.

    Raises
    ------
    TypeError if arg_to_check is not a str or list of str
    """
    # check that return_tags has the right type:
    if isinstance(arg_to_check, str):
        arg_to_check = [arg_to_check]
    if not isinstance(arg_to_check, list) or not all(
        isinstance(value, str) for value in arg_to_check
    ):
        raise TypeError(
            f"Error in all_objects!  Argument {arg_name} must be either\
             a str or list of str"
        )
    return arg_to_check


def _get_return_tags(object, return_tags):
    """Fetch a list of all tags for every_entry of all_objects.

    Parameters
    ----------
    object: BaseObject, an skpro object
    return_tags: list of str,
        names of tags to get values for the object

    Returns
    -------
    tags: a tuple with all the objects values for all tags in return tags.
        a value is None if it is not a valid tag for the object provided.
    """
    tags = tuple(object.get_class_tag(tag) for tag in return_tags)
    return tags


def _check_tag_cond(object, filter_tags=None, as_dataframe=True):
    """Check whether object satisfies filter_tags condition.

    Parameters
    ----------
    object: BaseObject, an skpro object
    filter_tags: dict of (str or list of str), default=None
        subsets the returned objects as follows:
            each key/value pair is statement in "and"/conjunction
                key is tag name to sub-set on
                value str or list of string are tag values
                condition is "key must be equal to value, or in set(value)"
    as_dataframe: bool, default=False
                if False, return is as described below;
                if True, return is converted into a pandas.DataFrame for pretty
                display

    Returns
    -------
    cond_sat: bool, whether object satisfies condition in filter_tags
    """
    if not isinstance(filter_tags, dict):
        raise TypeError("filter_tags must be a dict")

    cond_sat = True

    for key, value in filter_tags.items():
        if not isinstance(value, list):
            value = [value]
        cond_sat = cond_sat and object.get_class_tag(key) in set(value)

    return cond_sat


def all_tags(
    object_types=None,
    as_dataframe=False,
):
    """Get a list of all tags from skpro.

    Retrieves tags directly from `_tags`, offers filtering functionality.

    Parameters
    ----------
    object_types: string, list of string, optional (default=None)
        Which kind of objects should be returned.
        If None, no filter is applied and all objects are returned.
    as_dataframe: bool, optional (default=False)
        if False, return is as described below;
        if True, return is converted into a pandas.DataFrame for pretty display

    Returns
    -------
    tags: list of tuples (a, b, c, d),
        in alphabetical order by a
        a : string - name of the tag as used in the _tags dictionary
        b : string - name of the scitype this tag applies to
                    must be in _base_classes.BASE_CLASS_SCITYPE_LIST
        c : string - expected type of the tag value
            should be one of:
                "bool" - valid values are True/False
                "int" - valid values are all integers
                "str" - valid values are all strings
                ("str", list_of_string) - any string in list_of_string is valid
                ("list", list_of_string) - any individual string and sub-list is valid
        d : string - plain English description of the tag
    """

    def is_tag_for_type(tag, object_types):
        tag_types = tag[1]
        tag_types = _check_list_of_str_or_error(tag_types, "tag_types")

        if isinstance(object_types, str):
            object_types = [object_types]

        tag_types = set(tag_types)
        object_types = set(object_types)
        is_valid_tag_for_type = len(tag_types.intersection(object_types)) > 0

        return is_valid_tag_for_type

    all_tags = OBJECT_TAG_REGISTER

    if object_types:
        # checking, but not using the return since that is classes, not strings
        _check_object_types(object_types)
        all_tags = [tag for tag in all_tags if is_tag_for_type(tag, object_types)]

    all_tags = sorted(all_tags, key=itemgetter(0))

    # convert to pd.DataFrame if as_dataframe=True
    if as_dataframe:
        columns = ["name", "scitype", "type", "description"]
        all_tags = pd.DataFrame(all_tags, columns=columns)

    return all_tags


def _check_object_types(object_types):
    """Return list of classes corresponding to type strings."""
    object_types = deepcopy(object_types)

    if not isinstance(object_types, list):
        object_types = [object_types]  # make iterable

    def _get_err_msg(object_type):
        return (
            f"Parameter `object_type` must be None, a string or a list of "
            f"strings. Valid string values are: "
            f"{tuple(VALID_OBJECT_TYPE_STRINGS)}, but found: "
            f"{repr(object_type)}"
        )

    for object_type in object_types:
        if not isinstance(object_type, (type, str)):
            raise ValueError(
                "Please specify `object_types` as a list of str or " "types."
            )
        if isinstance(object_type, str):
            if object_type not in VALID_OBJECT_TYPE_STRINGS:
                raise ValueError(_get_err_msg(object_type))
    return object_types



"""
Simple Bayesian Linear Regressor.

Bayesian Linear Regression defined with user-specified priors or defaults for slopes,
intercept, and noise; implemented using the pymc backend.
"""

# copyright: skpro developers
__author__ = ["meraldoantonio"]

from skbase.utils.dependencies import _check_soft_dependencies

from skpro.regression.base import BaseProbaRegressor


class BayesianLinearRegressor(BaseProbaRegressor):
    """
    Bayesian Linear Regression class with MCMC sampling.

    Defined with user-specified priors or defaults for slopes, intercept,
    and noise; implemented using the pymc backend.

    Parameters
    ----------
    prior_config : Dictionary, optional
        Dictionary of priors
        Class-default defined by default_prior_config method.
    sampler_config : Dictionary, optional
        Dictionary of parameters that initialise sampler configuration.
        Class-default defined by default_sampler_config method.

    Example
    -------
    >>> from skpro.regression.bayesian import BayesianLinearRegressor
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>> from sklearn.model_selection import train_test_split  # doctest: +SKIP
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)  # doctest: +SKIP

    >>> bayes_model = BayesianLinearRegressor()  # doctest: +SKIP
    >>> bayes_model.fit(X_train, y_train)  # doctest: +SKIP
    >>> y_test_pred_proba = bayes_model.predict_proba(X_test)  # doctest: +SKIP
    >>> y_test_pred = bayes_model.predict(X_test)  # doctest: +SKIP
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["meraldoantonio"],
        "python_version": ">=3.10",
        "python_dependencies": [
            "pymc",
            "pymc_marketing",
            "arviz>=0.18.0",
        ],
        # estimator tags
        # --------------
        "capability:multioutput": False,  # can the estimator handle multi-output data?
        "capability:missing": True,  # can the estimator handle missing data?
        "X_inner_mtype": "pd_DataFrame_Table",  # type seen in internal _fit, _predict
        "y_inner_mtype": "pd_DataFrame_Table",  # type seen in internal _fit
    }

    def __init__(self, prior_config=None, sampler_config=None):
        if sampler_config is None:
            sampler_config = {}
        if prior_config is None:
            prior_config = {}  # configuration for priors
        self.sampler_config = {**self.default_sampler_config, **sampler_config}
        self.prior_config = {**self.default_prior_config, **prior_config}
        self.model = None  # generated during fitting
        self.idata = None  # generated during fitting
        self._predict_done = False  # a flag indicating if a prediction has been done

        print(  # noqa: T201
            f"instantiated {self.__class__.__name__} with the following priors:"
        )

        for key, value in self.prior_config.items():
            print(f"  - {key}: {value}")  # noqa: T201

        super().__init__()

    @property
    def default_prior_config(self):
        """Return a dictionary of prior defaults."""
        from pymc_marketing.prior import Prior

        print(  # noqa: T201
            "The model assumes that the intercept and slopes are independent. \n\
            Modify the model if this assumption doesn't apply!"
        )
        default_prior_config = {
            "intercept": Prior(
                "Normal", mu=0, sigma=100
            ),  # Weakly informative normal prior with large sigma
            "slopes": Prior(
                "Normal", mu=0, sigma=100, dims=("pred_id",)
            ),  # Same for slopes
            "noise_var": Prior(
                "HalfCauchy", beta=5
            ),  # Weakly informative Half-Cauchy prior for noise variance
        }
        return default_prior_config

    @property
    def default_sampler_config(self):
        """Return a class default sampler configuration dictionary."""
        default_sampler_config = {
            "draws": 1000,
            "tune": 1000,
            "chains": 2,
            "target_accept": 0.95,
            "random_seed": 123,
            "progressbar": True,
        }
        return default_sampler_config

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        import warnings

        import pandas as pd
        import pymc as pm

        assert len(y.columns) == 1, "y must have only one column!"
        self._X = X
        self._y = y
        self._y_vals = y.values[
            :, 0
        ]  # we need a 1-dimensional array for compatibility with pymc

        # Model construction and posterior sampling
        with pm.Model(coords={"obs_id": X.index, "pred_id": X.columns}) as self.model:
            # Mutable data containers for X and y
            X_data = pm.Data("X", X, dims=("obs_id", "pred_id"))
            y_data = pm.Data("y", self._y_vals, dims=("obs_id"))

            # Priors for model parameters, taken from self.prior_config
            self.intercept = self.prior_config["intercept"].create_variable("intercept")
            self.slopes = self.prior_config["slopes"].create_variable("slopes")
            self.noise_var = self.prior_config["noise_var"].create_variable("noise_var")
            self.noise = pm.Deterministic("noise", self.noise_var**0.5)

            # Expected value of the target variable
            self.mu = pm.Deterministic(
                "mu", self.intercept + pm.math.dot(X_data, self.slopes)
            )

            # Likelihood of observations
            y_obs = pm.Normal(  # noqa: F841
                "y_obs", mu=self.mu, sigma=self.noise, observed=y_data, dims=("obs_id")
            )

            # Constructing the posterior
            self.idata = pm.sample(**self.sampler_config)

        # Incorporation of training_data as a new group in self.idata
        training_data = pd.concat([X, y], axis=1)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
            )
            self.idata.add_groups(training_data=training_data.to_xarray())
        return self

    def visualize_model(self, **kwargs):
        """Use Graphviz to visualize the model flow."""
        _check_soft_dependencies(
            "graphviz", msg="You need to install Graphviz to use this method!"
        )
        import pymc as pm

        assert self._is_fitted, "You need to fit the model before visualizing it!"

        return pm.model_to_graphviz(self.model, **kwargs)

    def _sample_dataset(self, group_name, return_type=None):
        """
        General method to sample from a specified group in the idata object.

        Extracts samples from a specified group (e.g., 'prior') in the idata object and
        returns them in the required format

        Parameters
        ----------
        group_name : str
            The name of the group in the idata object to sample from (e.g., 'prior').

        return_type : str or None, optional (default=None)
            The format in which to return the sampled distributions.
            Accepted values are:
            - "xarray": Returns an xarray.Dataset
            - "numpy": Returns a dictionary of NumPy arrays
            - "dataframe": Returns a pandas DataFrame
            - "skpro": Returns an `Empirical` distribution from the skpro library.
            - None: Does not return any sampled data but performs the sampling
                    and updates the 'idata' attribute.

        Returns
        -------
        xarray.Dataset or dict or pd.DataFrame or skpro.distributions.Empirical or None

            The sampled distributions in the specified format,
            or None if return_type is None.
        """
        import pandas as pd

        # Validate the return_type
        assert return_type in [
            "xarray",
            "numpy",
            "dataframe",
            "skpro",
            None,
        ], "return_type must be one of 'xarray', 'numpy', 'dataframe', 'skpro', or None"

        # Validate that the group_name exists in idata
        assert hasattr(
            self.idata, group_name
        ), f"{group_name} group does not exist in the idata object."

        # Get the specified group from idata
        group = getattr(self.idata, group_name)
        # prediction-specific groups which focus on posterior predictive
        is_predictive = group_name in ["predictions", "posterior_predictive"]
        # as opposed to ["prior", "posterior"] which focus on prior/posterior

        if is_predictive:
            variables = ["y_obs"]
        else:
            variables = ["intercept", "slopes", "noise_var", "noise"]

        if return_type is None:
            return None
        elif return_type == "xarray":
            return group
        else:
            data_dict = {}

            for var in variables:
                # Check if the variable has a `pred_id` dimension
                if var in group and "pred_id" in group[var].dims:
                    # Iterate through each feature (e.g., 'feature1', 'feature2')
                    for feature in group[var].pred_id.values:
                        # Select the slope for the current feature and flatten it
                        feature_key = f"{var}_{feature}"
                        data_dict[feature_key] = (
                            group[var]
                            .sel(pred_id=feature)
                            .stack({"sample": ("chain", "draw")})
                            .values.squeeze()
                        )
                else:
                    if var in group:
                        data_dict[var] = (
                            group[var]
                            .stack({"sample": ("chain", "draw")})
                            .values.squeeze()
                        )

            if return_type == "numpy":
                return data_dict

            elif return_type == "dataframe":
                if is_predictive:
                    return pd.DataFrame(data_dict["y_obs"]).T
                else:
                    return pd.DataFrame(data_dict)

            elif return_type == "skpro":
                from skpro.distributions import Empirical

                if not is_predictive:
                    df = pd.DataFrame(data_dict)
                    reshaped_df = df.stack()
                    reshaped_df = reshaped_df.reset_index(name="value")
                    reshaped_df.set_index(["level_0", "level_1"], inplace=True)
                    reshaped_df.index.names = ["obs_id", "variable"]
                    return Empirical(spl=reshaped_df)
                else:
                    # Extract posterior predictive distributions as an xarray DataArray
                    pred_proba_xarray = group["y_obs"]

                    # Convert data to pd.DataFrame and format it appropriately for
                    # subsequent conversion into a skpro Empirical distribution
                    pred_proba_df = pred_proba_xarray.to_dataframe()
                    pred_proba_df = pred_proba_df.reset_index()

                    # Create a new 'sample_id' column by
                    # combining the 'chain' and 'draw' columns
                    pred_proba_df["sample_id"] = (
                        pred_proba_df["chain"] * self.sampler_config["draws"]
                        + pred_proba_df["draw"]
                    )
                    pred_proba_df = pred_proba_df[["obs_id", "sample_id", "y_obs"]]
                    pred_proba_df = pred_proba_df.rename(
                        columns={"y_obs": self._y.columns[0]}
                    )
                    pred_proba_df = pred_proba_df.set_index(["sample_id", "obs_id"])

                    # Convert data to skpro Empirical distribution
                    pred_proba_dist = Empirical(
                        spl=pred_proba_df, columns=self._y.columns
                    )
                    return pred_proba_dist

    def _get_dataset_summary(self, group_name, var_names=None, **kwargs):
        """
        Get the summary statistics of a specified group in the idata object.

        Parameters
        ----------
        group_name : str
            The name of the group in the idata object to summarize (e.g., 'prior').

        var_names : list, optional (default=None)
            A list of variable names to include in the summary.
            If None, all variables in the group are included.

        **kwargs :
            Additional keyword arguments to pass to `arviz.summary`.

        Returns
        -------
        az.data.inference_data.Summary
            The summary statistics for the specified group and variables.
        """
        import arviz as az

        # Check if the specified group exists in the idata object
        if group_name not in self.idata.groups():
            if group_name == "prior":
                self.sample_prior()
            elif group_name == "posterior":
                self.sample_posterior()
            else:
                raise ValueError(
                    f"Group '{group_name}' does not exist in the idata object."
                )

        # Get the summary statistics with optional kwargs
        return az.summary(
            getattr(self.idata, group_name), var_names=var_names, **kwargs
        )

    def sample_prior(self, return_type=None):
        """
        Sample from the prior distributions.

        Samples from the prior distributions and returns
        them in the required format

        If return_type is None, the method updates the 'idata' attribute
        by adding the 'prior' group but does not return any samples.

        return_type : str or None, optional (default=None)
            The format in which to return the sampled distributions.
            Accepted values are:
            - "xarray": Returns an xarray.Dataset
            - "numpy": Returns a dictionary of NumPy arrays
            - "dataframe": Returns a pandas DataFrame
            - "skpro": Returns an `Empirical` distribution from the skpro library.
            - None: Does not return any sampled data but performs the sampling
                    and updates the 'idata' attribute.

        Returns
        -------
        xarray.Dataset or dict or pd.DataFrame or skpro.distributions.Empirical or None
            The sampled distributions in the specified format,
            or None if return_type is None.
        """
        import pymc as pm

        assert (
            self.is_fitted
        ), "Model needs to be fitted before you can sample from prior"

        with self.model:
            # if we've previously used the model for prediction,
            # we need to reset the reference of 'X' to X used for training
            if self._predict_done:
                pm.set_data(
                    {"X": self._X},
                    coords={"obs_id": self._X.index, "pred_id": self._X.columns},
                )
            self.idata.extend(
                pm.sample_prior_predictive(
                    samples=self.sampler_config["draws"],
                    random_seed=self.sampler_config["random_seed"],
                )
            )  # todo: the keyword 'samples' will be changed to 'draws'
            # in pymc 5.16

        return self._sample_dataset(
            group_name="prior",
            return_type=return_type,
        )

    def get_prior_summary(self, **kwargs):
        """
        Get the summary statistics of prior distributions.

        Parameters
        ----------
        **kwargs :
            Additional keyword arguments to pass to `arviz.summary`.

        Returns
        -------
        az.data.inference_data.Summary
            The summary statistics for the prior distributions.
        """
        return self._get_dataset_summary(
            group_name="prior",
            var_names=["intercept", "slopes", "noise_var", "noise"],
            **kwargs,
        )

    def sample_posterior(self, return_type=None):
        """
        Sample from the posterior distributions.

        Samples from the posterior distributions and returns
        them in the required format

        If return_type is None, the method updates the 'idata' attribute
        by adding the 'posterior' group but does not return any samples.

        return_type : str or None, optional (default="xarray")
            The format in which to return the sampled distributions.
            Accepted values are:
            - "xarray": Returns an xarray.Dataset
            - "numpy": Returns a dictionary of NumPy arrays
            - "dataframe": Returns a pandas DataFrame
            - "skpro": Returns an `Empirical` distribution from the skpro library.
            - None: Does not return any sampled data but performs the sampling
                    and updates the 'idata' attribute.

        Returns
        -------
        xarray.Dataset or dict or pd.DataFrame or skpro.distributions.Empirical or None
            The sampled distributions in the specified format,
            or None if return_type is None.
        """
        assert (
            self.is_fitted
        ), "The model must be fitted before posterior can be returned."
        return self._sample_dataset(
            group_name="posterior",
            return_type=return_type,
        )

    def get_posterior_summary(self, **kwargs):
        """
        Get the summary statistics of the posterior distributions.

        Parameters
        ----------
        **kwargs :
            Additional keyword arguments to pass to `arviz.summary`.

        Returns
        -------
        az.data.inference_data.Summary
            The summary statistics for the posterior distributions.
        """
        return self._get_dataset_summary(
            group_name="posterior",
            var_names=["intercept", "slopes", "noise_var", "noise"],
            **kwargs,
        )

    def sample_in_sample_posterior_predictive(self, return_type=None):
        """Perform in-sample predictions and sample from it."""
        import pymc as pm

        with self.model:
            # if we've previously used the model for prediction,
            # we need to reset the reference of 'X' to X_train (i.e. self._X)
            if self._predict_done:
                pm.set_data(
                    {"X": self._X},
                    coords={"obs_id": self._X.index, "pred_id": self._X.columns},
                )
            self.idata.extend(
                pm.sample_posterior_predictive(self.idata, predictions=False)
            )

        return self._sample_dataset(
            group_name="posterior_predictive", return_type=return_type
        )

    def plot_ppc(self, **kwargs):
        """Plot the posterior predictive check."""
        import arviz as az

        if "posterior_predictive" not in self.idata:
            self.sample_in_sample_posterior_predictive()

        return az.plot_ppc(self.idata, **kwargs)

    def _predict_proba(self, X):
        """
        Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        pred_proba_dist : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        import pymc as pm

        with self.model:
            if "predictions" in self.idata.groups():
                del self.idata.predictions

            # Set the X to be the new 'X' variable and then sample posterior predictive
            pm.set_data({"X": X}, coords={"obs_id": X.index, "pred_id": X.columns})
            self.idata.extend(
                pm.sample_posterior_predictive(
                    self.idata,
                    predictions=True,
                )
            )
            self._predict_done = True  # a flag indicating prediction has been done

        return self._sample_dataset(group_name="predictions", return_type="skpro")

    # todo: return default parameters, so that a test instance can be created
    # required for automated unit and integration testing of estimator
    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from pymc_marketing.prior import Prior

        params1 = {}
        params2 = {"prior_config": {"intercept": Prior("Normal", mu=0, sigma=10)}}

        return [params1, params2]
